import itertools
from ploomber.static_analysis.parser.tokens import (Assignment, Name, Operator,
                                                    Null)


class Parser:
    """
    The current implementation is very simple and does not even
    build an AST. It just parses one statement at a time. Fine for our
    purposes with some limitations.

    Parameters
    ----------
    tokens : list
        Tokens, obtained from a Lexer
    """
    def __init__(self, tokens):
        self.tokens = tokens
        self.pos = 0

    @property
    def current_token(self):
        return self.tokens[self.pos]

    @property
    def next_token(self):
        return self.tokens[self.pos + 1]

    def get_tail(self, exclude_next=True):
        return self.tokens[self.pos + 1 + int(exclude_next):]

    def parse(self):
        """The current implementation can only parse one expression at a time
        """
        if not isinstance(self.current_token, Name):
            raise SyntaxError('First token must be a valid name')

        if not isinstance(self.next_token, Assignment):
            raise SyntaxError('Second token must be an assignment')

        return Expression(self.current_token, self.next_token,
                          build_node(self.get_tail()))


def get_slicer(elements, size):
    elements = iter(elements)

    while True:
        slice_ = list(itertools.islice(elements, size))

        if not slice_:
            return

        if len(slice_) == size - 1:

            slice_.append(None)

        yield slice_


class Node:
    pass


class Expression(Node):
    def __init__(self, left, op, right):
        self.left = left
        self.op = op
        self.right = right

    def __repr__(self):
        return '{} {} {}'.format(self.left, self.op, self.right)


class ListNode(Node):
    def __init__(self, tokens):
        self.elements = [value for value, comma in get_slicer(tokens, size=2)]

    def to_python(self):
        return [e.value for e in self.elements]


class DictionaryNode(Node):
    def __init__(self, tokens):
        self.elements = [
            (key, value)
            for key, equal, value, comma in get_slicer(tokens, size=4)
        ]

    def to_python(self):
        return {key.value: value.value for key, value in self.elements}


def build_node(tokens):
    if tokens[0] == Operator('list'):
        elements = tokens[2:-1]

        if isinstance(elements[0], Name):
            return DictionaryNode(elements)
        else:
            return ListNode(elements)

    elif tokens[0] == Null():
        return tokens[0]
    else:
        raise SyntaxError('Variables should be assigned to values of type '
                          'list (e.g. list("a", "b"))  or NULL')



from ploomber.sources.sources import (SQLScriptSource, SQLQuerySource,
                                      GenericSource, FileSource, EmptySource)
from ploomber.sources.notebooksource import NotebookSource
from ploomber.sources.pythoncallablesource import PythonCallableSource

__all__ = [
    'PythonCallableSource', 'SQLScriptSource', 'SQLQuerySource',
    'GenericSource', 'FileSource', 'NotebookSource', 'EmptySource'
]



import os
import shlex
import pdb
import tempfile
import subprocess
from subprocess import PIPE
from pathlib import Path
from importlib.util import find_spec
import warnings

import jupytext
import nbformat
import nbconvert
from nbconvert import ExporterNameError

# this was introduced in nbconvert 6.0
try:
    from nbconvert.exporters.webpdf import WebPDFExporter
except ModuleNotFoundError:
    WebPDFExporter = None

# papermill is importing a deprecated module from pyarrow
with warnings.catch_warnings():
    warnings.simplefilter('ignore', FutureWarning)
    import papermill as pm

try:
    # this is an optional dependency
    from pyppeteer import chromium_downloader
except ImportError:
    chromium_downloader = None

from ploomber.exceptions import TaskBuildError, TaskInitializationError
from ploomber.sources import NotebookSource
from ploomber.sources.notebooksource import _cleanup_rendered_nb
from ploomber.products import File, MetaProduct
from ploomber.tasks.abc import Task
from ploomber.util import requires, chdir_code
from ploomber.io import FileLoaderMixin, pretty_print
from ploomber.util._sys import _python_bin


# TODO: ensure that all places where we call this function are unit tested
def _write_text_utf_8(path, text):
    """Write text using UTF-8 encoding

    Notes
    -----
    In some systems, UTF-8 is not the default encoding (this happens on some
    Windows configurations, which have CP-1252 as default). This can break
    storing some files if the str to write contains characters that do not
    exist in the default encoding
    (https://github.com/ploomber/ploomber/issues/334)

    Under normal circumstances, it's better to use the default encoding since
    it provides better system compatibility, however, notebooks are always
    stored as UTF-8 so it's ok to store them as such.

    nbconvert always stores as UTF-8:
    https://github.com/jupyter/nbconvert/blob/5d57c16b655602f4705f9147f671f2cbaadaebca/nbconvert/writers/files.py#L126

    nbformat always reads as UTF-8:
    https://github.com/jupyter/nbformat/blob/df63593b64a15ee1c37b522973c39e8674f93c5b/nbformat/__init__.py#L125
    """
    Path(path).write_text(text, encoding='utf-8')


def _suggest_passing_product_dictionary():
    if Path('pipeline.yaml').is_file():
        return """

If you want to generate multiple products, pass a dictionary to \
product in your pipeline.yaml:

product:
  # nb must contain the path to the output notebook
  nb: products/output-notebook.ipynb
  # other outputs (optional)
  data: products/some-data.csv
  more-data: products/more-data.csv
"""
    else:
        return ('\nIf you want this task '
                'to generate multiple products, pass a dictionary '
                'to "product", with the path to the '
                'output notebook in the "nb" key (e.g. "output.ipynb") '
                'and any other output paths in other keys)')


def _check_exporter(exporter, path_to_output):
    """
    Validate if the user can use the selected exporter
    """
    if WebPDFExporter is not None and exporter is WebPDFExporter:
        pyppeteer_installed = find_spec('pyppeteer') is not None

        if not pyppeteer_installed:
            raise TaskInitializationError(
                'pyppeteer is required to use '
                'webpdf, install it '
                'with:\npip install "nbconvert[webpdf]"')
        else:
            if not chromium_downloader.check_chromium():
                chromium_downloader.download_chromium()

        if Path(path_to_output).suffix != '.pdf':
            raise TaskInitializationError(
                'Expected output to have '
                'extension .pdf when using the webpdf '
                f'exporter, got: {path_to_output}. Change the extension '
                'and try again')


class NotebookConverter:
    """
    Thin wrapper around nbconvert to provide a simple API to convert .ipynb
    files to different formats.

    Parameters
    ----------
    nbconvert_export_kwargs : dict, default=None
        Keyword arguments passed to ``nbconvert.export``

    Notes
    -----
    The exporter is searched at initialization time to raise an appropriate
    error during Task setup rather than raising it at Task runtime

    Handles cases where the output representation is text (e.g. HTML) or bytes
    (e.g. PDF)
    """
    # mapping of extensions to the corresponding nbconvert exporter
    EXTENSION2EXPORTER = {
        '.md': 'markdown',
        '.html': 'html',
        '.tex': 'latex',
        '.pdf': 'pdf',
        '.rst': 'rst',
        '.ipynb': 'ipynb',
    }

    def __init__(self,
                 path_to_output,
                 exporter_name=None,
                 nbconvert_export_kwargs=None):
        self._suffix = Path(path_to_output).suffix

        if exporter_name is None:

            # try to infer exporter name from the extension
            if not self._suffix:
                raise TaskInitializationError(
                    'Could not determine format for product '
                    f'{pretty_print.try_relative_path(path_to_output)!r} '
                    'because it has no extension. '
                    'Add a valid one '
                    f'({pretty_print.iterable(self.EXTENSION2EXPORTER)}) '
                    'or pass "nbconvert_exporter_name".' +
                    _suggest_passing_product_dictionary())

            if self._suffix in self.EXTENSION2EXPORTER:
                exporter_name = self.EXTENSION2EXPORTER[self._suffix]
            else:
                raise TaskInitializationError(
                    'Could not determine '
                    'format for product '
                    f'{pretty_print.try_relative_path(path_to_output)!r}. '
                    'Pass a valid extension '
                    f'({pretty_print.iterable(self.EXTENSION2EXPORTER)}) or '
                    'pass "nbconvert_exporter_name".' +
                    _suggest_passing_product_dictionary())

        self._exporter = self._get_exporter(exporter_name, path_to_output)

        self._path_to_output = path_to_output
        self._nbconvert_export_kwargs = nbconvert_export_kwargs or {}

    def convert(self):
        if self._exporter is None and self._nbconvert_export_kwargs:
            warnings.warn(
                f'Output {self._path_to_output!r} is a '
                'notebook file. nbconvert_export_kwargs '
                f'{self._nbconvert_export_kwargs!r} will be '
                'ignored since they only apply '
                'when exporting the notebook to other formats '
                'such as html. You may change the extension to apply '
                'the conversion parameters')

        if self._exporter is not None:
            self._from_ipynb(self._path_to_output, self._exporter,
                             self._nbconvert_export_kwargs)

    @staticmethod
    def _get_exporter(exporter_name, path_to_output):
        """
        Get function to convert notebook to another format using nbconvert,
        first. In some cases, the exporter name matches the file extension
        (e.g html) but other times it doesn't (e.g. slides), use
        `nbconvert.get_export_names()` to get available exporter_names

        Returns None if passed exported name is 'ipynb', raises ValueError
        if an exporter can't be located
        """
        extension2exporter_name = {'md': 'markdown'}

        # sometimes extension does not match with the exporter name, fix
        # if needed
        if exporter_name in extension2exporter_name:
            exporter_name = extension2exporter_name[exporter_name]

        if exporter_name == 'ipynb':
            exporter = None
        else:
            try:
                exporter = nbconvert.get_exporter(exporter_name)

            # nbconvert 5.6.1 raises ValueError, beginning in version 6,
            # it raises ExporterNameError. However the exception is defined
            # since 5.6.1 so we can safely import it
            except (ValueError, ExporterNameError):
                error = True
            else:
                error = False

            if error:
                names = nbconvert.get_export_names()
                raise TaskInitializationError(
                    f"{exporter_name!r} is not a "
                    "valid 'nbconvert_exporter_name' value. "
                    "Choose one from: "
                    f'{pretty_print.iterable(names)}')

        _check_exporter(exporter, path_to_output)

        return exporter

    @staticmethod
    def _from_ipynb(path_to_nb, exporter, nbconvert_export_kwargs):
        """
        Convert ipynb to another format
        """

        path = Path(path_to_nb)

        nb = nbformat.reads(path.read_text(), as_version=nbformat.NO_CONVERT)
        content, _ = nbconvert.export(exporter, nb, **nbconvert_export_kwargs)

        if isinstance(content, str):
            _write_text_utf_8(path, content)
        elif isinstance(content, bytes):
            path.write_bytes(content)
        else:
            raise TypeError('nbconvert returned a converted notebook with'
                            'unknown format, only text and binary objects '
                            'are supported')

        return content


class NotebookMixin(FileLoaderMixin):
    # expose the static_analysis attribute from the source, we need this
    # since we disable static_analysis when rendering the DAG in Jupyter

    @property
    def static_analysis(self):
        return self._source.static_analysis

    @static_analysis.setter
    def static_analysis(self, value):
        self._source.static_analysis = value

    def develop(self, app='notebook', args=None):
        """
        Opens the rendered notebook (with injected parameters) and adds a
        "debugging-settings" cell to the that changes directory to the current
        active directory. This will reflect conditions when callign
        `DAG.build()`. This modified notebook is saved in the same location as
        the source with a "-tmp" added to the filename. Changes to this
        notebook can be exported to the original notebook after the notebook
        process is shut down. The "injected-parameters" and
        "debugging-settings" cells are deleted before saving.

        Parameters
        ----------
        app : {'notebook', 'lab'}, default: 'notebook'
            Which Jupyter application to use
        args : str
            Extra parameters passed to the jupyter application

        Notes
        -----
        Be careful when developing tasks interacively. If the task has run
        successfully, you overwrite products but don't save the
        updated source code, your DAG will enter an inconsistent state where
        the metadata won't match the overwritten product.

        If you modify the source code and call develop again, the source
        code will be updated only if the ``hot_reload option`` is turned on.
        See :class:`ploomber.DAGConfigurator` for details.
        """
        # TODO: this code needs refactoring, should be a context manager
        # like the one we have for PythonCallable.develop that abstracts
        # the handling of the temporary notebook while editing

        apps = {'notebook', 'lab'}

        if app not in apps:
            raise ValueError('"app" must be one of {}, got: "{}"'.format(
                apps, app))

        if self.source.language != 'python':
            raise NotImplementedError(
                'develop is not implemented for "{}" '
                'notebooks, only python is supported'.format(
                    self.source.language))

        if self.source.loc is None:
            raise ValueError('Can only use develop in notebooks loaded '
                             'from files, not from str')

        nb = _read_rendered_notebook(self.source.nb_str_rendered)

        name = self.source.loc.name
        suffix = self.source.loc.suffix
        name_new = name.replace(suffix, '-tmp.ipynb')
        tmp = self.source.loc.with_name(name_new)
        content = nbformat.writes(nb, version=nbformat.NO_CONVERT)
        _write_text_utf_8(tmp, content)

        # open notebook with injected debugging cell
        try:
            subprocess.run(['jupyter', app, str(tmp)] +
                           shlex.split(args or ''),
                           check=True)
        except KeyboardInterrupt:
            print(f'Jupyter {app} application closed...')

        # read tmp file again, to see if the user made any changes
        content_new = Path(tmp).read_text()

        # maybe exclude changes in tmp cells?
        if content == content_new:
            print('No changes found...')
        else:
            # save changes
            if _save():
                nb = nbformat.reads(content_new,
                                    as_version=nbformat.NO_CONVERT)

                # remove injected-parameters and debugging-settings cells if
                # they exist
                _cleanup_rendered_nb(nb)

                # write back in the same format and original location
                ext_source = Path(self.source.loc).suffix[1:]
                print('Saving notebook to: ', self.source.loc)
                jupytext.write(nb, self.source.loc, fmt=ext_source)
            else:
                print('Not saving changes...')

        # remove tmp file
        Path(tmp).unlink()

    def debug(self, kind='ipdb'):
        """
        Opens the notebook (with injected parameters) in debug mode in a
        temporary location

        Parameters
        ----------
        kind : str, default='ipdb'
            Debugger to use, 'ipdb' to use line-by-line IPython debugger,
            'pdb' to use line-by-line Python debugger or 'pm' to to
            post-portem debugging using IPython

        Notes
        -----
        Be careful when debugging tasks. If the task has run
        successfully, you overwrite products but don't save the
        updated source code, your DAG will enter an inconsistent state where
        the metadata won't match the overwritten product.
        """
        if self.source.language != 'python':
            raise NotImplementedError(
                'debug is not implemented for "{}" '
                'notebooks, only python is supported'.format(
                    self.source.language))

        opts = {'ipdb', 'pdb', 'pm'}

        if kind not in opts:
            raise ValueError('kind must be one of {}'.format(opts))

        nb = _read_rendered_notebook(self.source.nb_str_rendered)
        fd, tmp_path = tempfile.mkstemp(suffix='.py')
        os.close(fd)
        code = jupytext.writes(nb, version=nbformat.NO_CONVERT, fmt='py')
        _write_text_utf_8(tmp_path, code)

        if kind == 'pm':
            # post-mortem debugging
            try:
                subprocess.run(['ipython', tmp_path, '--pdb'])
            finally:
                Path(tmp_path).unlink()
        else:
            if kind == 'ipdb':
                from IPython.terminal.debugger import TerminalPdb, Pdb
                code = compile(source=code, filename=tmp_path, mode='exec')

                try:
                    # this seems to only work in a Terminal
                    debugger = TerminalPdb()
                except Exception:
                    # this works in a Jupyter notebook
                    debugger = Pdb()

            elif kind == 'pdb':
                debugger = pdb

            try:
                debugger.run(code)
            finally:
                Path(tmp_path).unlink()


class NotebookRunner(NotebookMixin, Task):
    """
    Run a Jupyter notebook using papermill. Support several input formats
    via jupytext and several output formats via nbconvert

    Parameters
    ----------
    source: str or pathlib.Path
        Notebook source, if str, the content is interpreted as the actual
        notebook, if pathlib.Path, the content of the file is loaded. When
        loading from a str, ext_in must be passed
    product: ploomber.File
        The output file
    dag: ploomber.DAG
        A DAG to add this task to
    name: str, optional
        A str to indentify this task. Should not already exist in the dag
    params: dict, optional
        Notebook parameters. This are passed as the "parameters" argument
        to the papermill.execute_notebook function, by default, "product"
        and "upstream" are included
    papermill_params : dict, optional
        Other parameters passed to papermill.execute_notebook, defaults to None
    kernelspec_name: str, optional
        Kernelspec name to use, if the file extension provides with enough
        information to choose a kernel or the notebook already includes
        kernelspec data (in metadata.kernelspec), this is ignored, otherwise,
        the kernel is looked up using jupyter_client.kernelspec.get_kernel_spec
    nbconvert_exporter_name: str, optional
        Once the notebook is run, this parameter controls whether to export
        the notebook to a different parameter using the nbconvert package,
        it is not needed unless the extension cannot be used to infer the
        final output format, in which case the nbconvert.get_exporter is used.
    ext_in: str, optional
        Source extension. Required if loading from a str. If source is a
        ``pathlib.Path``, the extension from the file is used.
    nb_product_key: str, optional
        If the notebook is expected to generate other products, pass the key
        to identify the output notebook (i.e. if product is a list with 3
        ploomber.File, pass the index pointing to the notebook path). If the
        only output is the notebook itself, this parameter is not needed
    static_analysis : ('disabled', 'regular', 'strict'), default='regular'
        Check for various errors in the notebook. In 'regular' mode, it aborts
        execution if the notebook has syntax issues, or similar problems that
        would cause the code to break if executed. In 'strict' mode, it
        performs the same checks but raises an isse before starting execution
        of any task, furthermore, it verifies that the parameters cell and
        the params passed to the notebook match, thus, making the notebook
        behave like a function with a signature.
    nbconvert_export_kwargs : dict
        Keyword arguments to pass to the ``nbconvert.export`` function (this is
        only used if exporting the output ipynb notebook to another format).
        You can use this, for example, to hide code cells using the
        exclude_input parameter. See ``nbconvert`` documentation for details.
        Ignored if the product is file with .ipynb extension.
    local_execution : bool, optional
        Change working directory to be the parent of the notebook's source.
        Defaults to False. This resembles the default behavior when
        running notebooks interactively via `jupyter notebook`


    Examples
    --------
    >>> from pathlib import Path
    >>> from ploomber import DAG
    >>> from ploomber.tasks import NotebookRunner
    >>> from ploomber.products import File
    >>> dag = DAG()
    >>> # do not include input code (only cell's output)
    >>> NotebookRunner(Path('nb.ipynb'), File('out-1.html'), dag=dag,
    ...                nbconvert_export_kwargs={'exclude_input': True},
    ...                name=1) # doctest: +SKIP
    >>> # Selectively remove cells with the tag "remove"
    >>> config = {'TagRemovePreprocessor': {'remove_cell_tags': ('remove',)},
    ...        'HTMLExporter':
    ...         {'preprocessors':
    ...    ['nbconvert.preprocessors.TagRemovePreprocessor']}} # doctest: +SKIP
    >>> NotebookRunner(Path('nb.ipynb'), File('out-2.html'), dag=dag,
    ...                nbconvert_export_kwargs={'config': config},
    ...                name=2) # doctest: +SKIP
    >>> dag.build() # doctest: +SKIP

    Notes
    -----
    nbconvert's documentation:
    https://nbconvert.readthedocs.io/en/latest/config_options.html#preprocessor-options
    """
    PRODUCT_CLASSES_ALLOWED = (File, )

    @requires(['jupyter', 'papermill', 'jupytext'], 'NotebookRunner')
    def __init__(self,
                 source,
                 product,
                 dag,
                 name=None,
                 params=None,
                 papermill_params=None,
                 kernelspec_name=None,
                 nbconvert_exporter_name=None,
                 ext_in=None,
                 nb_product_key='nb',
                 static_analysis='regular',
                 nbconvert_export_kwargs=None,
                 local_execution=False,
                 check_if_kernel_installed=True):
        self.papermill_params = papermill_params or {}
        self.nbconvert_export_kwargs = nbconvert_export_kwargs or {}
        self.kernelspec_name = kernelspec_name
        self.nbconvert_exporter_name = nbconvert_exporter_name
        self.ext_in = ext_in
        self.nb_product_key = nb_product_key
        self.local_execution = local_execution
        self.check_if_kernel_installed = check_if_kernel_installed

        if 'cwd' in self.papermill_params and self.local_execution:
            raise KeyError('If local_execution is set to True, "cwd" should '
                           'not appear in papermill_params, as such '
                           'parameter will be set by the task itself')

        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = NotebookRunner._init_source(source, kwargs, ext_in,
                                                   kernelspec_name,
                                                   static_analysis,
                                                   check_if_kernel_installed)
        super().__init__(product, dag, name, params)

        if isinstance(self.product, MetaProduct):
            if self.product.get(nb_product_key) is None:
                raise TaskInitializationError(
                    f"Missing key '{nb_product_key}' in "
                    f"product: {(str(self.product))!r}. "
                    f"{nb_product_key!r} must contain "
                    "the path to the output notebook.")

        if isinstance(self.product, MetaProduct):
            product_nb = (
                self.product[self.nb_product_key]._identifier.best_repr(
                    shorten=False))
        else:
            product_nb = self.product._identifier.best_repr(shorten=False)

        self._converter = NotebookConverter(product_nb,
                                            nbconvert_exporter_name,
                                            nbconvert_export_kwargs)

    @staticmethod
    def _init_source(source,
                     kwargs,
                     ext_in=None,
                     kernelspec_name=None,
                     static_analysis='regular',
                     check_if_kernel_installed=False):
        return NotebookSource(
            source,
            ext_in=ext_in,
            kernelspec_name=kernelspec_name,
            static_analysis=static_analysis,
            check_if_kernel_installed=check_if_kernel_installed,
            **kwargs)

    def run(self):
        # regular mode: raise but not check signature
        # strict mode: called at render time
        if self.static_analysis == 'regular':
            self.source._check_notebook(raise_=True, check_signature=False)

        if isinstance(self.product, MetaProduct):
            path_to_out = Path(str(self.product[self.nb_product_key]))
        else:
            path_to_out = Path(str(self.product))

        # we will run the notebook with this extension, regardless of the
        # user's choice, if any error happens, this will allow them to debug
        # we will change the extension after the notebook runs successfully
        path_to_out_ipynb = path_to_out.with_suffix('.ipynb')

        fd, tmp = tempfile.mkstemp('.ipynb')
        os.close(fd)

        tmp = Path(tmp)
        _write_text_utf_8(tmp, self.source.nb_str_rendered)

        if self.local_execution:
            self.papermill_params['cwd'] = str(self.source.loc.parent)

        # create parent folders if they don't exist
        Path(path_to_out_ipynb).parent.mkdir(parents=True, exist_ok=True)

        try:
            # no need to pass parameters, they are already there
            pm.execute_notebook(str(tmp), str(path_to_out_ipynb),
                                **self.papermill_params)
        except Exception as e:
            raise TaskBuildError(
                'Error when executing task'
                f' {self.name!r}. Partially'
                f' executed notebook available at {str(path_to_out_ipynb)}'
            ) from e
        finally:
            tmp.unlink()

        # on windows, Path.rename will throw an error if the file exists
        if path_to_out_ipynb != path_to_out and path_to_out.is_file():
            path_to_out.unlink()

        path_to_out_ipynb.rename(path_to_out)
        self._converter.convert()


class ScriptRunner(NotebookMixin, Task):
    """
    Similar to NotebookRunner, except it uses python to run the code,
    instead of papermill, hence, it doesn't generate an output notebook. But
    it also works by injecting a cell into the source code. Source can be
    a .py script or an .ipynb notebook. Does not support magics.
    """
    @requires(['jupyter', 'jupytext'], 'ScriptRunner')
    def __init__(self,
                 source,
                 product,
                 dag,
                 name=None,
                 params=None,
                 ext_in=None,
                 static_analysis='regular',
                 local_execution=False):
        self.ext_in = ext_in
        self.local_execution = local_execution

        kwargs = dict(hot_reload=dag._params.hot_reload)
        self._source = ScriptRunner._init_source(source, kwargs, ext_in,
                                                 static_analysis)
        super().__init__(product, dag, name, params)

    @staticmethod
    def _init_source(source, kwargs, ext_in=None, static_analysis='regular'):
        return NotebookSource(source,
                              ext_in=ext_in,
                              static_analysis=static_analysis,
                              kernelspec_name=None,
                              check_if_kernel_installed=False,
                              **kwargs)

    def run(self):
        # regular mode: raise but not check signature
        # strict mode: called at render time
        if self.static_analysis == 'regular':
            self.source._check_notebook(raise_=True, check_signature=False)

        if self.local_execution:
            cwd = str(self.source.loc.parent)
        else:
            cwd = None

        fd, tmp = tempfile.mkstemp('.py')
        os.close(fd)

        code = '\n\n'.join([
            c['source'] for c in self.source.nb_obj_rendered.cells
            if c['cell_type'] == 'code'
        ])

        tmp = Path(tmp)
        tmp.write_text(code)

        if self.source.language == 'python':
            interpreter = _python_bin()
        elif self.source.language == 'r':
            interpreter = 'Rscript'
        else:
            raise ValueError(
                'ScriptRunner only works with Python and R scripts')

        try:
            _run_script_in_subprocess(interpreter, tmp, cwd)
        except Exception as e:
            raise TaskBuildError('Error when executing task'
                                 f' {self.name!r}.') from e
        finally:
            tmp.unlink()


def _run_script_in_subprocess(interpreter, path, cwd):
    res = subprocess.run([interpreter, str(path)], cwd=cwd, stderr=PIPE)

    if res.returncode:
        stderr = res.stderr.decode()

        if 'SyntaxError' in stderr:
            stderr += ('(Note: IPython magics are not supported in '
                       'ScriptRunner, remove them or use the regular '
                       'NotebookRunner)')

        raise RuntimeError('Error while executing ScriptRunner:\n' f'{stderr}')


def _read_rendered_notebook(nb_str):
    """
    Read rendered notebook and inject cell with debugging settings
    """
    # add debug cells
    nb = nbformat.reads(nb_str, as_version=nbformat.NO_CONVERT)
    nbformat_v = nbformat.versions[nb.nbformat]

    source = """
# Debugging settings (this cell will be removed before saving)
# change the current working directory to directory of the session that
# invoked the jupyter app to make relative paths work
import os
{}
""".format(chdir_code(Path('.').resolve()))

    cell = nbformat_v.new_code_cell(source,
                                    metadata={'tags': ['debugging-settings']})
    nb.cells.insert(0, cell)

    return nb


def _save():
    res = input('Notebook changed, do you want to save changes '
                'in the original location? (injected parameters '
                'and debugging cells will be removed before '
                'saving). Enter "no" to skip saving changes, '
                'anything else will be interpreted as "yes": ')
    return res != 'no'



"""
Extractors read "upstream" and "product" variables defined in sources
"""
import abc


class Extractor(abc.ABC):
    def __init__(self, code):
        self.code = code

    @abc.abstractmethod
    def extract_upstream(self):
        pass

    @abc.abstractmethod
    def extract_product(self):
        pass


class NotebookExtractor(Extractor):
    def __init__(self, parameters_cell):
        self.parameters_cell = parameters_cell

        if self.parameters_cell is None:
            raise ValueError('Notebook does not have a cell with the tag '
                             '"parameters"')



from ploomber.static_analysis.parser.parser import Parser
from ploomber.static_analysis.parser.rlexer import RLexer
from ploomber.static_analysis.abc import NotebookExtractor


def naive_parsing(code, var_name):
    """
    Our current R parser can only deal with a single statement (one line)
    at a time. So we parse on a per-line basis and look for the variable
    we want, this will be replaced for a more efficient implementation
    once we improve the parser
    """
    for code in code.split('\n'):
        # ignore empty lines and comments
        if code != '' and not code.startswith('#'):
            parser = Parser(list(RLexer(code)))
            exp = parser.parse()

            if exp.left.value == var_name:
                return exp.right.to_python()


class RNotebookExtractor(NotebookExtractor):

    def extract_upstream(self):
        parsed = naive_parsing(self.parameters_cell, 'upstream')
        return parsed if not parsed else set(parsed)

    def extract_product(self):
        return naive_parsing(self.parameters_cell, 'product')



"""
A mapping object with text and HTML representations
"""
from functools import reduce
from warnings import warn
from textwrap import TextWrapper
from copy import deepcopy
import shutil
from collections.abc import Mapping, Iterable
from tabulate import tabulate

from ploomber.util.util import isiterable_not_str

_BETWEEN_COLUMN_WIDTH = 2


class Row:
    """A class to represent a dictionary as a table row

    Parameters
    ----------
    mapping
        Maps column names to a single value

    Examples
    --------
    >>> from ploomber.table import Row
    >>> row = Row({'a': 'some value', 'b': 'another value'})
    """
    def __init__(self, mapping):
        if not isinstance(mapping, Mapping):
            raise TypeError('Rows must be initialized with mappings')

        self._set_mapping(mapping)

    def __str__(self):
        return self._str

    def __repr__(self):
        return str(self)

    def _repr_html_(self):
        return self._html

    def __getitem__(self, key):
        if isiterable_not_str(key):
            return Row({k: self._mapping[k] for k in key})
        else:
            return self._mapping[key]

    def __setitem__(self, key, value):
        self._mapping[key] = value

    def __eq__(self, other):
        return self._mapping == other

    @property
    def columns(self):
        return tuple(self._mapping.keys())

    def _set_mapping(self, mapping):
        self._mapping = mapping
        self._str = tabulate([self._mapping],
                             headers='keys',
                             tablefmt='simple')
        self._html = tabulate([self._mapping], headers='keys', tablefmt='html')


class Table:
    """A collection of rows

    Parameters
    ----------
    rows
        List of Row objects
    column_width
        Maximum column width, if None, no trimming happens, otherwise values
        are converted to string and trimmed. If 'auto' width is determined
        based on terminal sized (using ``shutil.get_terminal_size()``)
    """
    # Columns to exclude from wrapping
    EXCLUDE_WRAP = None

    def __init__(self, rows, column_width='auto'):
        self.column_width = column_width

        if isinstance(rows, list):
            self._values = rows2columns(rows)
        else:
            self._values = rows

        self._values = self.data_preprocessing(self._values)

        self._str = None
        self._html = None

    def __str__(self):
        if self._str is None or self.column_width == 'auto':
            values = wrap_table_dict(self._values, self.column_width,
                                     self.EXCLUDE_WRAP)
            self._str = tabulate(values, headers='keys', tablefmt='simple')

        return self._str

    def __repr__(self):
        return str(self)

    def _repr_html_(self):
        if self._html is None or self.column_width == 'auto':
            values = wrap_table_dict(self._values, self.column_width,
                                     self.EXCLUDE_WRAP)

            self._html = tabulate(values, headers='keys', tablefmt='html')

        return self._html

    def __getitem__(self, key):
        if isiterable_not_str(key):
            return Table({k: v
                          for k, v in self.values.items() if k in key},
                         column_width=self.column_width)
        else:
            return self.values[key]

    def __iter__(self):
        for col in self.values:
            yield col

    def __len__(self):
        return len(self._values.keys())

    def __eq__(self, other):
        return self.values == other

    def data_preprocessing(self, values):
        return values

    def to_format(self, fmt):
        values = wrap_table_dict(self.values, self.column_width,
                                 self.EXCLUDE_WRAP)
        return tabulate(values, headers='keys', tablefmt=fmt)

    def to_pandas(self):
        import pandas as pd
        return pd.DataFrame(self.values)

    def to_dict(self):
        return deepcopy(self.values)

    @property
    def values(self):
        return self._values

    @classmethod
    def from_dicts(cls, dicts, complete_keys=False):
        if complete_keys:
            keys_all = reduce(lambda a, b: set(a) | set(b), dicts)
            default = {k: '' for k in keys_all}
            return cls([Row({**default, **d}) for d in dicts])
        else:
            return cls([Row(d) for d in dicts])


class TaskReport(Row):
    EXCLUDE_WRAP = ['Ran?', 'Elapsed (s)']

    @classmethod
    def with_data(cls, name, ran, elapsed):
        return cls({'name': name, 'Ran?': ran, 'Elapsed (s)': elapsed})

    @classmethod
    def empty_with_name(cls, name):
        return cls.with_data(name, False, 0)


class BuildReport(Table):
    """A Table that adds a columns for checking task elapsed time
    """
    EXCLUDE_WRAP = ['Ran?', 'Elapsed (s)', 'Percentage']

    def data_preprocessing(self, values):
        """Create a build report from several tasks
        """
        # in case the pipeline has no tasks...
        elapsed = values.get('Elapsed (s)', [])

        total = sum(elapsed)

        def compute_pct(elapsed, total):
            if not elapsed:
                return 0
            else:
                return 100 * elapsed / total

        values['Percentage'] = [compute_pct(r, total) for r in elapsed]

        return values


def rows2columns(rows):
    """Convert [{key: value}, {key: value2}] to [{key: [value, value2]}]
    """
    if not len(rows):
        return {}

    cols_combinations = set(tuple(sorted(row.columns)) for row in rows)

    if len(cols_combinations) > 1:
        raise KeyError('All rows should have the same columns, got: '
                       '{}'.format(cols_combinations))

    columns = rows[0].columns

    return {col: [row[col] for row in rows] for col in columns}


def wrap_table_dict(table_dict, column_width, exclude):
    """Wraps a columns to take at most column_width characters

    Parameters
    ----------
    column_width : int, 'auto' or None
        Width per column. Splits evenly if 'auto', does not wrap if None
    exclude : list
        Exclude columns from wrapping (show them in a single line)
    """
    exclude = exclude or []

    if column_width is None:
        return table_dict

    if column_width == 'auto':
        column_width = calculate_wrapping(
            table_dict,
            do_not_wrap=exclude,
            width_total=shutil.get_terminal_size().columns)

    # NOTE: the output of this algorithm may return a table that does not use
    # between 0 and {column - 1} characters. We could always take all the
    # space available if we refactor and do not keep column_width fixed for
    # all columns
    wrapper = TextWrapper(width=column_width,
                          break_long_words=True,
                          break_on_hyphens=True)

    return apply_wrapping(table_dict, wrapper, exclude=exclude)


def separator_width(header_length, max_value_length):
    """
    Calculates the width of the '---' line that separates header from content
    """
    n_value_extra = header_length - max_value_length

    if n_value_extra >= -2:
        return header_length + 2
    else:
        return max_value_length


def width_required_for_column(header, values):
    """
    Spaced needed to display column in a single line, accounts for the two
    extra characters that the tabulate package adds to the header when the
    content is too short
    """
    values_max = -1 if not values else max(len(str(v)) for v in values)
    return max(values_max, separator_width(len(header), values_max))


def calculate_wrapping(table_dict, do_not_wrap, width_total):
    """
    Determines the column width by keeping some columns unwrapped (show all
    rows, including the header in a single line) and distributing the
    remaining space evenly. Accounts for the betwee-column spacing.
    """
    # space required to display a given column on a single column
    width_required = {
        header: width_required_for_column(header, values)
        for header, values in table_dict.items()
    }

    # TODO: pass set(table_dict) instead of table_dict
    column_width = _calculate_wrapping(table_dict, do_not_wrap, width_total,
                                       width_required)

    return column_width


def _calculate_wrapping(table_dict, do_not_wrap, width_total, width_required):
    width_offset = 0

    # how much space we are already taking by not wrapping columns in
    # do_not_wrap
    for col_name in do_not_wrap:
        # edge case: key might not exist if the dag is empty col_name
        # is added during Tabble.data_preprocessing (e.g.,) the "Ran?"" column
        # in a build report
        if col_name in width_required:
            width_offset += width_required[col_name]

    # how much we have left - takes into account the bweteen-column spacing
    # of two characters
    width_remaining = (width_total - width_offset -
                       len(do_not_wrap) * _BETWEEN_COLUMN_WIDTH)

    cols_to_wrap = set(table_dict) - set(do_not_wrap)

    # there should be at least one column to wrap to continue
    if not cols_to_wrap:
        return width_total

    # split remaining space evenly in the rest of the cols
    column_width = equal_column_width(n_cols=len(cols_to_wrap),
                                      width_total=width_remaining)

    # check if we have short columns. this means we can give more space
    # to the others
    short = {
        col
        for col in cols_to_wrap if width_required[col] <= column_width
    }

    # if there is a (strict) subset columns that can display in a single line
    # with less than column_width, call again. When short has all columns
    # already, just return whatever number we have, there are no more columns
    # to distribute space
    if short and short < set(table_dict):
        return _calculate_wrapping(table_dict,
                                   do_not_wrap=do_not_wrap + list(short),
                                   width_total=width_total,
                                   width_required=width_required)
    else:
        return column_width


def equal_column_width(n_cols, width_total):
    """
    Max column width if splitting width_total equally among n_cols. Note
    that before computing column width, a quantity is substracted to account
    for required for spacing between columns
    """
    if not n_cols:
        raise ValueError('n_cols must be >0')

    offset = (n_cols - 1) * _BETWEEN_COLUMN_WIDTH
    width_remaining = width_total - offset
    width_column = int(width_remaining / n_cols)

    # degenerate case: not even a single space to display. Return width of
    # 1 but show a warning, since the table will be illegible
    if width_column < 1:
        warn(f'Not enough space to display {n_cols} columns with '
             f'a width of {width_total}. Using a column width of 1')
        return 1

    return width_column


def apply_wrapping(table_dict, wrapper, exclude=None):
    """
    Wrap text using a wrapper, excluding columns in exclude
    """
    exclude = exclude or []

    return dict(
        apply_wrapping_to_column(header, values, exclude, wrapper)
        for header, values in table_dict.items())


def apply_wrapping_to_column(header, values, exclude, wrapper):
    # TODO: test this
    if header in exclude:
        return header, values

    # wrapping values is simple, apply the wrapper directly
    values_wrapped = wrap_elementwise(values, wrapper)

    # wrapping the header has a detail: if there isn't enough space
    # for the 2 character offset that tabulate adds to the header, the
    # column will end up taking more space than expected, we don't want that.
    # wrap the header a bit more if necessary. Clip to 0 since this can be
    # negative with large headers
    offset = max(wrapper.width - len(header), 0)

    if offset >= 2:
        header_wrapped = wrap_elementwise(header, wrapper)
    else:
        _wrapper = TextWrapper(width=wrapper.width - (2 - offset),
                               break_long_words=True,
                               break_on_hyphens=True)
        header_wrapped = wrap_elementwise(header, _wrapper)

    return header_wrapped, values_wrapped


def wrap_elementwise(value, wrapper):
    """Apply wrap if str (elementwise if iterable of str)
    """
    if isinstance(value, Iterable) and not isinstance(value, str):
        return [wrapper.fill(str(v)) for v in value]
    else:
        return wrapper.fill(str(value))



"""
Environment management
"""
from ploomber.env.envdict import EnvDict

# TODO: add defaults functionality if defined in {module}/env.defaults.yaml


class Env:
    """Return the current environment

    NOTE: this API is experimental and subject to change, it is recommmended
    to use @with_env and @load_env decorators instead

    Env provides a clean and consistent way of managing environment and
    configuration settings. Its simplest usage provides access to settings
    specified via an `env.yaml`.

    Settings managed by Env are intended to be runtime constant (they are NOT
    intended to be used as global variables). For example you might want
    to store database URIs. Storing sensitive information is discouraged as
    yaml files are plain text. Use `keyring` for that instead.

    All sections are optional, but if there is a path section, all values
    inside that section will be casted to pathlib.Path objects, expanduser()
    is applied so "~" can be used. Strings with a trailing "/" will be
    interpreted as directories and they will be created if they do not exist

    There are a few placeholders available:
    * {{user}} expands to the current user (by calling getpass.getuser())
    * {{version}} expands to module.__version__ if _module is defined
    * {{git}} expands to branch name if at the tip, otherwise to
    the current commit hash (_module has to be defined)

    Examples
    --------
    >>> from ploomber import Env
    >>> env = Env({'db': {'uri': 'my_uri'}, 'path': {'raw': '/path/to/raw'}})
    >>> env.db.uri
    'my_uri'
    >>> env.path.raw
    PosixPath('/path/to/raw')

    Notes
    -----
    Envs are intended to be short-lived, the recommended usage is to start and
    end them only during the execution of a function that builds a DAG by
    using the @with_env and @load_env decorators
    """
    __instance = None

    # just a variable to display in error messages so users know where
    # the env was initialized if they try to create a new one

    def __new__(cls, source=None):
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
            return cls.__instance
        else:
            raise RuntimeError('Cannot start environment, one has already '
                               'started: {}'.format(repr(cls.__instance)))

    def __init__(self, source='env.yaml'):
        """Start the environment

        Parameters
        ----------
        source: dict, pathlib.Path, str, optional
            If dict, loads it directly, if pathlib.Path or path, reads the file
            (assumes yaml format).

        Raises
        ------
        FileNotFoundError
            If source is None and an environment file cannot be found
            automatically
        RuntimeError
            If one environment has already started

        Returns
        -------
        ploomber.Env
            An environment object
        """
        if not isinstance(source, EnvDict):
            # try to initialize an EnvDict to perform validation, if any
            # errors occur, discard object
            try:
                source = EnvDict(source)
            except Exception:
                Env.__instance = None
                raise

        self._data = source
        self._fn_name = None

    @classmethod
    def _init_from_decorator(cls, source, fn_name):
        env = Env(source=source)
        env._fn_name = fn_name
        return env

    @classmethod
    def load(cls):
        if cls.__instance is None:
            raise RuntimeError('Env has not been set, run Env() before '
                               'running Env.load()')
        return cls.__instance

    @classmethod
    def end(cls):
        """
        End environment. Usage is discouraged, a single environment is expected
        to exist during the entire Python process lifespan to avoid
        inconsistencies, use it only if you have a very strong reason to
        """
        cls.__instance = None

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        s = 'Env({})'.format(str(self._data))

        if self._fn_name:
            s += ' (initialized in function: %s)' % self._fn_name

        if self._data.path_to_env:
            s += ' (from file: %s)' % str(self._data.path_to_env)

        return s

    def __dir__(self):
        return dir(self._data)

    @property
    def name(self):
        return self._data.name

    def __getattr__(self, key):
        return getattr(self._data, key)

    def __getitem__(self, key):
        return self._data[key]

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            raise RuntimeError('env is a read-only object')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        Env.end()



from copy import copy

from jinja2 import Template
import sqlparse


def strip(tokens):
    """
    Remove whitespace, comments and punctuation (;)
    """
    return [
        t for t in tokens if not t.is_whitespace
        and not isinstance(t, sqlparse.sql.Comment) and t.value != ';'
    ]


def strip_punctuation(tokens):
    return [t for t in tokens if t.value != ';']


def strip_comments(tokens):
    return [t for t in tokens if not isinstance(t, sqlparse.sql.Comment)]


def code_from_token_list(tokens):
    return ''.join([t.value for t in tokens])


def name_code_pair(f):
    # split to: ({name}, AS, {code})
    t_name, _, t_code = strip(f)
    name = t_name.value
    # ignore parenthesis
    code = code_from_token_list(t_code.tokens[1:-1])
    # NOTE: probably better to strip during SQLParser init?
    return name, sqlparse.format(code, strip_comments=True).strip()


def find_select(tokens):
    for idx, t in enumerate(tokens):
        # TODO: use .get_type instead of normalized
        if t.normalized == 'SELECT':
            return idx


def find_with_type(tokens, type_):
    for idx, t in enumerate(tokens):
        if t.get_type() == type_:
            return idx, t


def find_with_class(tokens, class_):
    for idx, t in enumerate(tokens):
        if isinstance(t, class_):
            return idx, t


def get_identifiers_and_select_from_create(create):
    res = find_with_class(create, class_=sqlparse.sql.IdentifierList)

    # when CREATE TABLE name has no parenthesis...
    if res:
        idx = find_select(create)
        select = code_from_token_list(strip_punctuation(create.tokens[idx:]))
        return strip(res[1]), select

    # when there are parenthesis
    _, identifier = find_with_class(create, class_=sqlparse.sql.Identifier)
    # last one is the parenthesis object that encloses the identifier list
    parenthesis = identifier.tokens[-1]
    # ignore parenthesis tokens at the beginning and end
    _, id_list = find_with_class(parenthesis,
                                 class_=sqlparse.sql.IdentifierList)

    idx = find_select(parenthesis)
    # ignore last token, it's the closing parenthesis
    select = code_from_token_list(strip_punctuation(
        parenthesis.tokens[idx:-1]))

    return strip(id_list), select


def find_identifiers_list(sql):
    # find the first of those (identifier happens when there is a single def)
    id_list = find_with_class(sql,
                              class_=(sqlparse.sql.Identifier,
                                      sqlparse.sql.IdentifierList))

    # if one identifier, wrap it as a list of length 1
    if isinstance(id_list[1], sqlparse.sql.Identifier):
        return [id_list[1]]
    else:
        return id_list[1]


def get_identifiers_and_select_from_with_select(with_select):
    defs = find_identifiers_list(with_select)

    idx = find_select(with_select)
    select = code_from_token_list(strip_punctuation(with_select.tokens[idx:]))

    return strip(defs), select


class SQLParser:
    """Parse a SQL script with format:

    CREATE [TABLE|VIEW] something  AS (
        WITH step_a AS (
            SELECT * FROM a
        ), step_b AS (
            SELECT * FROM B
        ), ...

        SELECT * FROM ...
    )

    CREATE [TABLE|VIEW] is optional.

    If there is more than one statement, the first CREATE [TABLE|VIEW] is used.
    Get individual SELECT statements using parser['step_a']

    Attributes
    ----------
    mapping : dict
        Contains the extracted subqueries
    """
    def __init__(self, sql):
        if sql is not None:
            # get the first create table statement
            out = find_with_type(sqlparse.parse(sql), type_='CREATE')

            if out is not None:
                # there is a create statement
                _, create = out

                # find the list of identifiers and last select statement
                ids, select = get_identifiers_and_select_from_create(create)
            else:
                # no create statement, look for the WITH ... SELECT statement
                _, with_select = find_with_type(sqlparse.parse(sql),
                                                type_='SELECT')
                ids, select = get_identifiers_and_select_from_with_select(
                    with_select)

            # this should be
            # {identifier} AS ( SELECT ... )
            # then punctiation ","
            # and repeat...
            functions = [id_ for idx, id_ in enumerate(ids) if not idx % 2]

            functions_t = [name_code_pair(f) for f in functions]

            m = {t[0]: t[1] for t in functions_t}

            m['_select'] = select

            self.mapping = m
        else:
            self.mapping = dict()

    def __getitem__(self, key):
        """Same as .until(key)
        """
        return self.until(key)

    def _ipython_key_completions_(self):
        return list(self)

    def __setitem__(self, key, value):
        self.mapping[key] = value

    def __iter__(self):
        for e in self.mapping:
            yield e

    def __len__(self):
        return len(self.mapping)

    def __repr__(self):
        return f'{type(self).__name__} with keys: {list(self.mapping)!r}'

    def until(self, key, select=None, limit=20, parse=True):
        """
        Generate with statements until the one with the given identifier.
        Adding a final select statement which can be customized with the
        "select" parameter, if "select" is None, the last select statement
        is used with a LIMIT statement. If you pass a custom select statement,
        you can include the {{key}} placeholder which is replaced by the key
        argument.
        """
        pairs = []

        for a_key in self.mapping:
            pairs.append((a_key, self.mapping[a_key]))

            if a_key == key:
                break

        if key == '_select':
            _, select = pairs.pop(-1)
        else:
            if select is None:
                select = f'SELECT * FROM {pairs[-1][0]}'

                if limit:
                    select += f' LIMIT {limit}'
            else:
                select = Template(select).render(key=key)

        sql = Template("""
WITH {%- for id,code in pairs -%}{{',' if not loop.first else '' }} {{id}} AS (
    {{code}}
){% endfor %}
{{select}}""").render(pairs=pairs, select=select)

        return sql if not parse else type(self)(sql)

    @classmethod
    def _with_mapping(cls, mapping):
        obj = cls(sql=None)
        obj.mapping = mapping
        return obj

    def insert_first(self, key, select, inplace=False):
        mapping = {key: select, **self.mapping}

        if inplace:
            self.mapping = mapping
            return self
        else:
            return type(self)._with_mapping(mapping)

    def insert_last(self, select, inplace=False):
        m = copy(self.mapping)
        m['last'] = m.pop('_select')
        mapping = {**m, '_select': select}

        if inplace:
            self.mapping = mapping
            return self
        else:
            return type(self)._with_mapping(mapping)

    def replace_last(self, select, inplace=False):
        mapping = copy(self.mapping)
        mapping['_select'] = select

        if inplace:
            self.mapping = mapping
            return self
        else:
            return type(self)._with_mapping(mapping)

    def __str__(self):
        """Short for .until(key='_select')
        """
        return self.to_str(select=None, limit=None)

    def to_str(self, select=None, limit=20):
        """Short for .until(key=None, *args, **kwargs)
        """
        return self.until(key='_select',
                          select=select,
                          limit=limit,
                          parse=False)



from ploomber.clients.client import Client
from ploomber.clients.db import DBAPIClient, SQLAlchemyClient
from ploomber.clients.shell import ShellClient, RemoteShellClient
from ploomber.clients.storage.gcloud import GCloudStorageClient
from ploomber.clients.storage.local import LocalStorageClient
from ploomber.clients.storage.aws import S3Client

__all__ = [
    'Client',
    'DBAPIClient',
    'SQLAlchemyClient',
    'ShellClient',
    'RemoteShellClient',
    'GCloudStorageClient',
    'LocalStorageClient',
    'S3Client',
]



"""
Tokens, given the limited scope our parser, we can reuse them for several
languages
"""


class Token:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.value == other.value

    def __repr__(self):
        return '{}("{}")'.format(type(self).__name__, self.value)

    def to_python(self):
        return self.value


class Operator(Token):
    pass


class Name(Token):
    pass


class Integer(Token):
    pass


class Null(Token):
    def __init__(self):
        self.value = None


class String(Token):
    pass


# FIXME: remove, just use Operator
class BinaryOperator(Token):
    pass


class Assignment(Token):
    pass





def [[function_name]](product):
    """Add description here
    """
    [% if extract_upstream -%]
    # extract_upstream=True in your pipeline.yaml file, if this task has
    # dependencies, add an "upstream" parameter to the function
    # and declare them here (e.g. upstream['some_task'])
    [% else -%]
    # extract_upstream=False in your pipeline.yaml file, if this task has
    # dependencies, add an "upstream" parameter to the function and declare
    # them in the YAML spec and refence them here
    [% endif -%]
    [% if extract_product -%]
    # extract_product=True in your pipeline.yaml file, declare a "product"
    # variable inside this function
    [% endif -%]
    pass



"""
Tools for quality assurance

This module contains utilities for performing checks in the DAGs source code
to look for bad practices. These is complementary to tools such as flake8.
Given that the DAG object has a lot of information, it can provide more
details than static analyzers can
"""
from ploomber.qa.dagqualitychecker import DAGQualityChecker

__all__ = ['DAGQualityChecker']



from ploomber.cli.parsers import CustomParser
from ploomber.cli.io import cli_endpoint
from ploomber.telemetry import telemetry


@cli_endpoint
@telemetry.log_call('status')
def main():
    parser = CustomParser(description='Show pipeline status',
                          prog='ploomber status')
    with parser:
        # this command has no static args
        pass

    dag, _ = parser.load_from_entry_point_arg()

    print(dag.status())



from ploomber.cli.parsers import CustomParser
from ploomber.cli.io import cli_endpoint
from ploomber.telemetry import telemetry
from ploomber.cloud import api

# TODO: we are just smoke testing this, we need to improve the tests
# (check the appropriate functions are called)


def _task_cli(accept_task_id=False):
    parser = CustomParser(description='Build tasks', prog='ploomber task')

    with parser:
        parser.add_argument('task_name')
        parser.add_argument('--source',
                            '-s',
                            help='Print task source code',
                            action='store_true')
        parser.add_argument(
            '--build',
            '-b',
            help='Build task (default if no other option is passed)',
            action='store_true')
        parser.add_argument('--force',
                            '-f',
                            help='Force task build (ignore up-to-date status)',
                            action='store_true')
        parser.add_argument('--status',
                            '-st',
                            help='Get task status',
                            action='store_true')
        parser.add_argument('--on-finish',
                            '-of',
                            help='Only execute on_finish hook',
                            action='store_true')

        if accept_task_id:
            parser.add_argument('--task-id')

    dag, args = parser.load_from_entry_point_arg()

    dag.render()
    task = dag[args.task_name]

    if args.source:
        print(task.source)

    if args.status:
        print(task.status())

    if args.on_finish:
        task._run_on_finish()

    # task if build by default, but when --source or --status are passed,
    # the --build flag is required
    no_flags = not any((args.build, args.status, args.source, args.on_finish))

    if no_flags or args.build:
        try:
            task.build(force=args.force)
        except Exception:
            if getattr(args, 'task_id', None):
                api.tasks_update(getattr(args, 'task_id'), 'failed')
            raise
        else:
            if getattr(args, 'task_id', None):
                api.tasks_update(getattr(args, 'task_id'), 'finished')


@cli_endpoint
@telemetry.log_call('task')
def main():
    _task_cli()


# TODO: test this
if __name__ == '__main__':
    _task_cli(accept_task_id=True)



"""
Utils for comparing source code
"""
import io
import tokenize
from difflib import Differ
import parso
import ast

try:
    import sqlparse
except ImportError:
    sqlparse = None

try:
    import autopep8
except ImportError:
    autopep8 = None
from ploomber.products.serializeparams import remove_non_serializable_top_keys


def normalize_null(code):
    return code


def normalize_sql(code):
    if not sqlparse:
        raise ImportError('sqlparse is required for normalizing SQL code')

    return None if code is None else sqlparse.format(code,
                                                     keyword_case='upper',
                                                     identifier_case='lower',
                                                     strip_comments=True,
                                                     reindent=True,
                                                     indent_with=4)


def _delete_python_comments(code):
    tokens = tokenize.generate_tokens(io.StringIO(code).readline)
    tokens = [(num, val) for num, val, _, _, _ in tokens
              if num != tokenize.COMMENT]
    return tokenize.untokenize(tokens)


def normalize_python(code):
    if code is None:
        return None

    try:
        ast.parse(code)

    except SyntaxError:
        pass

    else:
        code = _delete_python_comments(code)

        if not autopep8 or not parso:
            raise ImportError(
                'autopep8 and parso are required for normalizing '
                'Python code: pip install autopep8 parso')

        node = parso.parse(code).children[0]

        if node.type == 'decorated':
            node = node.children[-1]

        try:
            doc_node = node.get_doc_node()
        except Exception:
            # function without docstring...
            doc_node = None

        if doc_node is not None:
            code = code.replace('\n' + doc_node.get_code(), '')

        code = autopep8.fix_code(code)

    return code


def diff_strings(a, b):
    """Compute the diff between two strings
    """
    d = Differ()

    if a is None and b is None:
        return '[Both a and b are None]'

    out = ''

    if a is None:
        out += '[a is None]\n'
    elif b is None:
        out += '[a is None]\n'

    a = '' if a is None else a
    b = '' if b is None else b

    result = d.compare(a.splitlines(keepends=True),
                       b.splitlines(keepends=True))
    out += ''.join(result)

    return out


class CodeDiffer:
    NORMALIZERS = {
        None: normalize_null,
        'py': normalize_python,
        'sql': normalize_sql
    }

    def is_different(self, a, b, a_params, b_params, extension=None):
        """
        Compares code and params to determine if it's changed. Ignores top-keys
        in a_params or b_params if they're no JSON serializable.

        Parameters
        ----------
        a : str
            Code to compare

        b : str
            Code to compare

        a_params : dict
            Params passed to a

        b_params : dict
            Params passed to b

        extension : str, default=None
            Code extension. Used to normalize code to prevent changes such
            as whitespace to trigger false positives. Normalization only
            available for .py and .sql, other languages are compared as is

        Returns
        -------
        result : bool
            True if code is different (different code or params),
            False if they are the same (same code and params)

        diff : str
            A diff view of the differences

        Notes
        -----
        Params comparison is ignored if either a_params or b_params is None
        """
        # TODO: this can be more efficient. ie only compare source code
        # if params are the same and only get diff if result is True
        normalizer = self._get_normalizer(extension)
        a_norm = normalizer(a)
        b_norm = normalizer(b)

        if a_params is None or b_params is None:
            outdated_params = False
        else:
            a_params_ = remove_non_serializable_top_keys(a_params)
            b_params_ = remove_non_serializable_top_keys(b_params)
            outdated_params = (a_params_ != b_params_)

        result = outdated_params or (a_norm != b_norm)
        # TODO: improve diff view, also show a params diff view. probably
        # we need to normalize them first (maybe using pprint?) then take
        # the diff
        diff = self.get_diff(a_norm, b_norm, normalize=False)

        return result, diff

    def get_diff(self, a, b, extension=None, normalize=True):
        """Get the diff view
        """
        # TODO: remove normalize param, we are already normalizing in the
        # is_different method
        if normalize:
            normalizer = self._get_normalizer(extension)

            a = normalizer(a)
            b = normalizer(b)

        diff = diff_strings(a, b)

        if extension is not None:
            diff = '[Code was normalized]\n' + diff

        return diff

    def _get_normalizer(self, extension):
        """Get the normalizer function for a given extension
        """
        if extension in self.NORMALIZERS:
            return self.NORMALIZERS[extension]
        else:
            return normalize_null



"""
Clients that communicate with databases
"""
try:
    import sqlalchemy
except ImportError:
    sqlalchemy = None

from ploomber.clients.client import Client
from ploomber.util import requires

import re
from pathlib import Path


def code_split(code, token=';'):
    only_whitespace = re.compile(r'^\s*$')

    for part in code.split(token):
        if not re.match(only_whitespace, part):
            yield part


def _get_url_obj(uri):
    if isinstance(uri, str):
        return sqlalchemy.engine.make_url(uri)
    elif isinstance(uri, sqlalchemy.engine.url.URL):
        return uri
    else:
        raise TypeError('SQLAlchemyClient must be initialized with a '
                        'string or a sqlalchemy.engine.url.URL object, got '
                        f'{type(uri).__name__}')


class DBAPIClient(Client):
    """A client for a PEP 249 compliant client library

    Parameters
    ----------
    connect_fn : callable
        The function to use to open the connection

    connect_kwargs : dict
        Keyword arguments to pass to connect_fn

    split_source : str, optional
        Some database drivers do not support multiple commands in a single
        execute statement. Use this optiion to split commands by a given
        character (e.g. ';') and send them one at a time. Defaults to
        None (no splitting)

    Examples
    --------
    >>> from ploomber.clients import DBAPIClient
    >>> import sqlite3
    >>> client = DBAPIClient(sqlite3.connect, dict(database='my.db'))
    """
    def __init__(self, connect_fn, connect_kwargs, split_source=None):
        super().__init__()
        self.connect_fn = connect_fn
        self.connect_kwargs = connect_kwargs
        self.split_source = split_source

        # there is no open connection by default
        self._connection = None

    @property
    def connection(self):
        """Return a connection, open one if there isn't any
        """
        # if there isn't an open connection, open one...
        if self._connection is None:
            self._connection = self.connect_fn(**self.connect_kwargs)

        return self._connection

    def cursor(self):
        return self.connection.cursor()

    def execute(self, code):
        """Execute code with the existing connection
        """
        cur = self.connection.cursor()

        if self.split_source:
            for command in code_split(code, token=self.split_source):
                cur.execute(command)
        else:
            cur.execute(code)

        self.connection.commit()
        cur.close()

    def close(self):
        """Close connection if there is an active one
        """
        if self._connection is not None:
            self._connection.close()

    def __getstate__(self):
        state = super().__getstate__()
        state['_connection'] = None
        return state


class SQLAlchemyClient(Client):
    """Client for connecting with any SQLAlchemy supported database

    Parameters
    ----------
    uri: str or sqlalchemy.engine.url.URL
        URI to pass to sqlalchemy.create_engine or URL object created using
        sqlalchemy.engine.url.URL.create

    split_source : str, optional
        Some database drivers do not support multiple commands in a single
        execute statement. Use this optiion to split commands by a given
        character (e.g. ';') and send them one at a time. Defaults to
        'default', which splits by ';' if using SQLite database,
        but does not perform any splitting with other databases. If None,
        it will never split, a string value is interpreted as the token
        to use for splitting statements regardless of the database type

    create_engine_kwargs : dict, optional
        Keyword arguments to pass to ``sqlalchemy.create_engine``

    Notes
    -----
    SQLite client does not support sending more than one command at a time,
    if using such backend code will be split and several calls to the db
    will be performed.

    Examples
    --------
    >>> from ploomber.clients import SQLAlchemyClient
    >>> client = SQLAlchemyClient('sqlite:///my_db.db')
    >>> import sqlalchemy
    >>> url = sqlalchemy.engine.url.URL.create(drivername='sqlite',
    ...                                        database='my_db.db')
    >>> client = SQLAlchemyClient(url)
    """
    split_source_mapping = {'sqlite': ';'}

    @requires(['sqlalchemy'], 'SQLAlchemyClient')
    def __init__(self, uri, split_source='default', create_engine_kwargs=None):
        super().__init__()
        self._uri = uri

        self._url_obj = _get_url_obj(uri)
        self._uri_safe = repr(self._url_obj)

        self._create_engine_kwargs = create_engine_kwargs or dict()
        self.flavor = self._url_obj.drivername
        self._engine = None
        self.split_source = split_source

        if self.flavor == "sqlite" and self._url_obj.database:
            parent_folder = Path(self._url_obj.database).parent
            parent_folder.mkdir(exist_ok=True, parents=True)

        self._connection = None

    @property
    def connection(self):
        """Return a connection from the pool
        """
        # we have to keep this reference here,
        # if we just return self.engine.raw_connection(),
        # any cursor from that connection will fail
        # doing: engine.raw_connection().cursor().execute('') fails!
        if self._connection is None:
            self._connection = self.engine.raw_connection()

        # if a task or product calls client.connection.close(), we have to
        # re-open the connection
        if not self._connection.is_valid:
            self._connection = self.engine.raw_connection()

        return self._connection

    def cursor(self):
        return self.connection.cursor()

    def execute(self, code):
        cur = self.connection.cursor()

        # if split_source is default, and there's a default token defined,
        # use it
        if (self.split_source == 'default'
                and self.flavor in self.split_source_mapping):
            token = self.split_source_mapping[self.flavor]
            for command in code_split(code, token=token):
                cur.execute(command)
        # otherwise interpret split_source as the token
        elif self.split_source:
            for command in code_split(code, token=self.split_source):
                cur.execute(command)
        # if no split_source, execute all at once
        else:
            cur.execute(code)

        self.connection.commit()
        cur.close()

    def close(self):
        """Closes all connections
        """
        if self._engine is not None:
            self._engine.dispose()
            self._engine = None

        if self._connection is not None:
            self._connection.close()

    @property
    def engine(self):
        """Returns a SQLAlchemy engine
        """
        if self._engine is None:
            self._engine = sqlalchemy.create_engine(
                self._uri, **self._create_engine_kwargs)

        return self._engine

    def __getstate__(self):
        state = super().__getstate__()
        state['_engine'] = None
        state['_connection'] = None
        return state

    def __str__(self):
        return self._uri_safe

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, self._uri_safe)



import subprocess


class CommanderTester:
    """Object to mock calls to ploomber.io._commander.Commander

    Parameters
    ----------
    run : list of tuples
        List of tuples, where each tuple represents the argument passed to
        subprocess.check_call. Whenever the mocked object calls, CommandTester
        checks if the passed arg is in the run list, if so, it will
        call subprocess.{check_call, check_output} depending on the Commander's
        capture_output argument value

    return_value : dict
        A mapping of tuples -> values. Whenever the mocked object calls,
        CommandTester checks if the passed arg is a key in this dictionary,
        if so, it returns the value corresponding to such key, instead
        of calling subprocess.{check_call, check_output}

    Examples
    --------
    >>> from unittest.mock import Mock
    >>> from ploomber.io import _commander
    >>> tester = CommanderTester(run=[('python', 'script.py')],
    ...                       return_value={('python', '--version'): b'3.10'})
    >>> subprocess_mock = Mock()
    >>> # we need to mock check_all and check_output since Commander uses both
    >>> subprocess_mock.check_call.side_effect = tester
    >>> subprocess_mock.check_output.side_effect = tester
    >>> monkeypatch.setattr(_commander, 'subprocess',
    ...     subprocess_mock) # doctest: +SKIP
    """
    def __init__(self, run=None, return_value=None):
        self._run = run or []
        self._return_value = return_value or dict()
        self._calls = []

    def __call__(self, cmd):
        self._calls.append(cmd)

        if cmd in self._run:
            return subprocess.check_call(cmd)
        elif cmd in self._return_value:
            return self._return_value[cmd]

    @property
    def calls(self):
        return self._calls



from copy import copy
import warnings
from ploomber.env.expand import iterate_nested_dict


def is_json_serializable(obj):
    # https://github.com/python/cpython/blob/dea59cf88adf5d20812edda330e085a4695baba4/Lib/json/encoder.py#L73
    return isinstance(
        obj, (dict, list, tuple, str, int, float, bool)) or obj is None


def remove_non_serializable_top_keys(obj):
    """
    Remove top-level keys with unserializable objects, warning if necessary

    Parameters
    ----------
    d: a dictionary containing parameters
    """
    out = copy(obj)

    for (_, _, current_val, preffix) in iterate_nested_dict(obj):
        if not is_json_serializable(current_val):
            top_key = preffix[0]

            if top_key in out:
                del out[top_key]

            warnings.warn(f'Param {top_key!r} contains an unserializable '
                          f'object: {current_val!r}, it will be ignored. '
                          'Changes to it will not trigger task execution.')

    return out



import shutil
import json
import pickle
from pathlib import Path
from functools import wraps
from inspect import signature
from collections.abc import Mapping, Iterable

try:
    from joblib import dump as joblib_dump
except ModuleNotFoundError:
    joblib_dump = None

try:
    from cloudpickle import dump as cloudpickle_dump
except ModuleNotFoundError:
    cloudpickle_dump = None

from ploomber.products import MetaProduct


def _str2txt(obj, product):
    if not isinstance(obj, str):
        raise TypeError(f'Error serializing product {product!r} to .txt '
                        'with default serializer: data must be str, '
                        f'not {type(obj).__name__}')

    Path(product).write_text(obj)


def _obj2json(obj, product):
    try:
        serialized = json.dumps(obj)
    except TypeError:
        error = True
    else:
        error = False

    if error:
        raise TypeError(f'Error serializing product {product!r} to .json with '
                        'default serializer: Object of type  '
                        f'{type(obj).__name__} is not '
                        'JSON serializable')

    Path(product).write_text(serialized)


def _df2csv(obj, product):
    if not hasattr(obj, 'to_csv'):
        raise TypeError(f'Error serializing product {product!r} to .csv '
                        'with default serializer: expected a pandas.DataFrame '
                        f'but got an object of type {type(obj).__name__}')

    obj.to_csv(product)


def _df2parquet(obj, product):
    if not hasattr(obj, 'to_parquet'):
        raise TypeError(f'Error serializing product {product!r} to .parquet '
                        'with default serializer: expected a pandas.DataFrame '
                        f'but got an object of type {type(obj).__name__}')

    obj.to_parquet(product)


_EXTERNAL = {
    False: None,
    True: pickle.dump,
    'pickle': pickle.dump,
    'joblib': joblib_dump,
    'cloudpickle': cloudpickle_dump,
}

_DEFAULTS = {
    '.txt': _str2txt,
    '.json': _obj2json,
    '.csv': _df2csv,
    '.parquet': _df2parquet,
}


def _extension_mapping_validate(extension_mapping, fn):
    if extension_mapping is not None:
        if not isinstance(extension_mapping, Mapping):
            raise TypeError(
                f'Invalid extension_mapping {extension_mapping!r} for '
                f'decorated function {fn.__name__!r}. Expected '
                'it to be a dictionary but got a '
                f'{type(extension_mapping).__name__}')

        invalid_keys = {
            k
            for k in extension_mapping.keys() if not k.startswith('.')
        }

        if invalid_keys:
            raise ValueError(
                f'Invalid extension_mapping {extension_mapping!r} for '
                f'decorated function {fn.__name__!r}. Expected '
                'keys to start with a dot (e.g., ".csv"). Invalid '
                f'keys found: {invalid_keys!r}')


def _build_extension_mapping_final(extension_mapping, defaults, fn,
                                   defaults_provided, name):
    defaults_keys = set(defaults_provided)

    if defaults:
        if not isinstance(defaults, Iterable) or isinstance(defaults, str):
            raise TypeError(f'Invalid defaults {defaults!r} for '
                            f'decorated function {fn.__name__!r}. Expected '
                            'it to be a list but got a '
                            f'{type(defaults).__name__}')

        passed_defaults = set(defaults)

        if extension_mapping:
            overlap = passed_defaults & set(extension_mapping)
            if overlap:
                raise ValueError(
                    f'Error when adding @{name} decorator '
                    f'to function {fn.__name__!r}: '
                    'Keys in \'extension_mapping\' and \'defaults\' must not '
                    f'overlap (overlapping keys: {overlap})')

        unexpected_defaults = passed_defaults - defaults_keys

        if unexpected_defaults:
            raise ValueError(
                f'Error when adding @{name} decorator '
                f'to function {fn.__name__!r}: unexpected values in '
                '"defaults" argument. Valid values are: '
                f'{defaults_keys}. Unexpected '
                f'values: {unexpected_defaults}')

        defaults_map = {
            k: v
            for k, v in defaults_provided.items() if k in defaults
        }
        extension_mapping_final = {**defaults_map, **(extension_mapping or {})}
    else:
        extension_mapping_final = extension_mapping

    _extension_mapping_validate(extension_mapping_final, fn)

    return extension_mapping_final


def serializer(extension_mapping=None,
               *,
               fallback=False,
               defaults=None,
               unpack=False):
    """Decorator for serializing functions

    Parameters
    ----------
    extension_mapping : dict, default=None
        An extension -> function mapping. Calling the decorated function with a
        File of a given extension will use the one in the mapping if it exists,
        e.g., {'.csv': to_csv, '.json': to_json}.

    fallback : bool or str, default=False
        Determines what method to use if extension_mapping does not match the
        product to serialize. Valid values are True (uses the pickle module),
        'joblib', and 'cloudpickle'. If you use any of the last two, the
        corresponding moduel must be installed. If this is enabled, the
        body of the decorated function is never executed. To turn it off
        pass False.

    defaults : list, default=None
        Built-in serializing functions to use. Must be a list with any
        combinations of values: '.txt', '.json', '.csv', '.parquet'. To save
        to .txt, the returned object must be a string, for .json it must be
        a json serializable object (e.g., a list or a dict), for .csv and
        .parquet it must be a pandas.DataFrame. If using .parquet, a parquet
        library must be installed (e.g., pyarrow). If extension_mapping
        and defaults contain overlapping keys, an error is raised

    unpack : bool, default=False
        If True, it treats every element in a dictionary as a different
        file, calling the serializing function one per (key, value) pair and
        using the key as filename.
    """
    def _serializer(fn):
        extension_mapping_final = _build_extension_mapping_final(
            extension_mapping, defaults, fn, _DEFAULTS, 'serializer')

        try:
            serializer_fallback = _EXTERNAL[fallback]
        except KeyError:
            error = True
        else:
            error = False

        if error:
            raise ValueError(f'Invalid fallback argument {fallback!r} '
                             f'in function {fn.__name__!r}. Must be one of '
                             "True, 'joblib', or 'cloudpickle'")

        if serializer_fallback is None and fallback in {
                'cloudpickle', 'joblib'
        }:
            raise ModuleNotFoundError(
                f'Error serializing with function {fn.__name__!r}. '
                f'{fallback} is not installed')

        n_params = len(signature(fn).parameters)
        if n_params != 2:
            raise TypeError(f'Expected serializer {fn.__name__!r} '
                            f'to take 2 arguments, but it takes {n_params!r}')

        @wraps(fn)
        def wrapper(obj, product):
            if isinstance(product, MetaProduct):
                _validate_obj(obj, product)

                for key, value in obj.items():
                    _serialize_product(value, product[key],
                                       extension_mapping_final, fallback,
                                       serializer_fallback, fn, unpack)
            else:
                _serialize_product(obj, product, extension_mapping_final,
                                   fallback, serializer_fallback, fn, unpack)

        return wrapper

    return _serializer


@serializer(fallback=True)
def serializer_pickle(obj, product):
    """A serializer that pickles everything
    """
    # this should never execute
    raise RuntimeError('Error when serializing with pickle module')


def _validate_obj(obj, product):
    if not isinstance(obj, Mapping):
        raise TypeError('Error serializing task: if task generates multiple '
                        f'products {product!r} the function must return '
                        f'a dictionary with the same keys (got {obj!r}, an '
                        f'object of type {type(obj).__name__}).')

    unexpected = set(obj) - set(product.products.products)
    missing = set(product.products.products) - set(obj)

    if missing or unexpected:
        error = ('Error serializing task: task generates products '
                 f'{product!r} but the function did not return a '
                 f'dictonary ({obj!r}) with valid keys. ')

        if missing:
            error += f'Missing keys: {missing!r}. '

        if unexpected:
            error += f'Unexpected keys: {unexpected!r}. '

        raise ValueError(error)


def _serialize_product(obj, product, extension_mapping, fallback,
                       serializer_fallback, fn, unpack):
    """
    Determine which function to use for serialization. Note that this
    function operates on single products. If the task generates multiple
    products, this function is called multiple times.
    """
    suffix = Path(product).suffix

    if unpack and isinstance(obj, Mapping):
        parent = Path(product)

        # if the directory exists, delete it, otherwise old files will
        # mix with the new ones
        if parent.is_dir():
            shutil.rmtree(product)
        # if it's a file, delete it as well
        elif parent.is_file():
            parent.unlink()

        parent.mkdir(exist_ok=True, parents=True)

        for filename, o in obj.items():
            out_path = _Path(product, filename)

            suffix_current = Path(filename).suffix
            serializer = _determine_serializer(suffix_current,
                                               extension_mapping, fallback,
                                               serializer_fallback, fn)
            serializer(o, out_path)
    else:
        serializer = _determine_serializer(suffix, extension_mapping, fallback,
                                           serializer_fallback, fn)
        serializer(obj, product)


def _make_serializer(fn):
    def _serialize(obj, product):
        with open(product, 'wb') as f:
            fn(obj, f)

    return _serialize


def _determine_serializer(suffix, extension_mapping, fallback,
                          serializer_fallback, fn):
    # if there is a serializer for the given extension, use it...
    if extension_mapping and suffix in extension_mapping:
        return extension_mapping[suffix]
    # no serializer for the given extension, check fallback...
    elif fallback:
        return _make_serializer(serializer_fallback)
    # otherwise call the function's body...
    else:
        return fn


def _Path(parent, filename):
    try:
        return Path(parent, filename)
    except TypeError:
        pass

    raise TypeError('Error creating output path from key with value '
                    f'{filename!r}: expected str, bytes or os.PathLike '
                    f'object, not {type(filename).__name__}')



"""
Data testing using pandas. This is intended to be used by the spec API, if
using the Python API, it's better and more flexible to write your own
functions
"""
from pathlib import Path
import pandas as pd


def load_product(product):
    extension = Path(str(product)).suffix[1:]

    loader = getattr(pd, 'read_'+extension, None)

    if not loader:
        raise ValueError('Reading files with extension "{}" is currently '
                         'unsupported')
    else:
        return loader(product)


def no_nulls_in_columns(cols, product):
    return not nulls_in_columns(cols, product)


def nulls_in_columns(cols, product):
    """Check if any column has NULL values, returns bool
    """
    df = load_product(product)
    return df.isna().values.sum() > 0


def distinct_values_in_column(col, product):
    """Get distinct values in a column, returns a set
    """
    df = load_product(product)
    return set(df[col].unique())


def duplicates_in_column(col, product):
    """Check if a column has duplicated values, returns bool
    """
    df = load_product(product)
    return (df[col].value_counts() > 1).sum() > 0


def no_duplicates_in_column(col, product):
    return not duplicates_in_column(col, product)


def range_in_column(col, product):
    """Get range for a column, returns a (min_value, max_value) tuple
    """
    df = load_product(product)
    return df[col].min(), df[col].max()



"""
As we started adding more features to DAG, I added a few more parameters to
the constructor, default values cover a lot of cases and most of the time
only a few parameters are actually modified. To prevent making the DAG API
unnecessarily complex, custom behavior will be provided via this object.

This is based in the essence pattern by by Andy Carlson
http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.81.1898&rep=rep1&type=pdf
"""
from copy import copy
from ploomber.dag.dag import DAG
from ploomber.dag.dagconfiguration import DAGConfiguration


class DAGConfigurator:
    """An object to customize DAG behavior

    Note: this API is experimental an subject to change

    To keep the DAG API clean, only the most important parameters are included
    in the constructor, the rest are accesible via a DAGConfigurator object

    Available parameters:

    outdated_by_code: whether source code differences make a task outdated
    cache_rendered_status: keep results from dag.render() whenever are needed
    again (e.g. when calling dag.build()) or compute it again every time.

    cache_rendered_status: If True, once the DAG is rendered, subsequent calls
    to render will not do anything (rendering is implicitely called in build,
    plot, status), otherwise it will always render again.

    hot_reload: Reload sources whenever they are updated

    Examples
    --------
    >>> from ploomber import DAGConfigurator
    >>> configurator = DAGConfigurator()
    >>> configurator.params.outdated_by_code = True
    >>> configurator.params.cache_rendered_status = False
    >>> configurator.params.hot_reload = True
    >>> dag = configurator.create()
    """
    def __init__(self, d=None):
        if d:
            self._params = DAGConfiguration.from_dict(d)
        else:
            self._params = DAGConfiguration()

    @property
    def params(self):
        return self._params

    def create(self, *args, **kwargs):
        """Return a DAG with the given parameters

        *args, **kwargs
            Parameters to pass to the DAG constructor
        """
        dag = DAG(*args, **kwargs)
        dag._params = copy(self.params)
        return dag

    def __setattr__(self, key, value):
        if key != '_params':
            raise AttributeError('Cannot assign attributes to DAGConfigurator,'
                                 ' use configurator.params.param_name = value')
        super().__setattr__(key, value)



"""
A Product specifies a persistent object in disk such as a file in the local
filesystem or an table in a database. Each Product is uniquely identified,
for example a file can be specified using a absolute path, a table can be
fully specified by specifying a database, a schema and a name. Names
are lazy evaluated, they can be built from templates
"""
import abc
import logging

from ploomber.products.metadata import Metadata
from ploomber.products._resources import process_resources


def _prepare_metadata(metadata):
    return metadata


class Product(abc.ABC):
    """
    Abstract class for all Products

    Attributes
    ----------
    prepare_metadata : callable
        A hook to execute before saving metadata, should include a "metadata"
        parameter and might include "product". "metadata" will be a
        dictionary with the metadata to save, it is not recommended to change
        any of the existing keys but additional key-value pairs might be
        included
    """

    # TODO: previously, File didn't have a client parameter but it does now,
    # it's best to include it here to simplify the constructors in the concrete
    # classes
    def __init__(self, identifier):
        self._identifier = self._init_identifier(identifier)

        if self._identifier is None:
            raise TypeError('_init_identifier must return a value, returned '
                            'None')

        self.task = None
        self.logger = logging.getLogger('{}.{}'.format(__name__,
                                                       type(self).__name__))

        self._outdated_data_dependencies_status = None
        self._outdated_code_dependency_status = None
        self._is_outdated_status = None
        # not all products have clients, but they should still have a client
        # property to keep the API consistent
        self._client = None
        self.metadata = Metadata(self)

        self.prepare_metadata = _prepare_metadata

    @property
    def task(self):
        if self._task is None:
            raise ValueError('This product has not been assigned to any Task')

        return self._task

    @property
    def client(self):
        return self._client

    @task.setter
    def task(self, value):
        self._task = value

    def render(self, params, **kwargs):
        """
        Render Product - this will render contents of Templates used as
        identifier for this Product, if a regular string was passed, this
        method has no effect
        """
        self._identifier.render(params, **kwargs)

    def _is_outdated(self, outdated_by_code=True):
        """
        Given current conditions, determine if the Task that holds this
        Product should be executed

        Returns
        -------
        bool
            True if the Task should execute
        """
        if self._is_outdated_status is None:
            self._is_outdated_status = self._check_is_outdated(
                outdated_by_code)

        return self._is_outdated_status

    def _is_remote_outdated(self, outdated_by_code):
        """
        Check if the remote version of this Product is outdated. Note that
        this is only valid for File products (which have their own)
        implementation. For non-File products (e.g., SQL products), there
        isn't a notion of "remote version", since they're already remote
        hence we simply use the existing implementation
        """
        return self._is_outdated(outdated_by_code=outdated_by_code)

    def _check_is_outdated(self, outdated_by_code):
        # check product...
        p_exists = self.exists()

        # check dependencies only if the product exists
        if p_exists:

            oudated_data = self._outdated_data_dependencies()
            outdated_code = (outdated_by_code
                             and self._outdated_code_dependency())
            run = oudated_data or outdated_code

            if run:
                self.logger.info(
                    'Task "%s" is outdated, it will be executed...',
                    self.task.name)
            else:
                self.logger.info(
                    'Task "%s" is up-to-date, it will be skipped...',
                    self.task.name)

            return run
        else:
            self.logger.info(
                'Product of task "%s" does not exist, it will be executed...',
                self.task.name)
            return True

    def _outdated_data_dependencies(self):
        """
        Determine if the product is outdated by checking upstream timestamps
        """

        if self._outdated_data_dependencies_status is not None:
            self.logger.debug(('Returning cached data dependencies status. '
                               'Outdated? %s'),
                              self._outdated_data_dependencies_status)
            return self._outdated_data_dependencies_status

        outdated = any([
            self._is_outdated_due_to_upstream(up.product)
            for up in self.task.upstream.values()
        ])

        self._outdated_data_dependencies_status = outdated

        self.logger.debug(('Finished checking data dependencies status. '
                           'Outdated? %s'),
                          self._outdated_data_dependencies_status)

        return self._outdated_data_dependencies_status

    def _is_outdated_due_to_upstream(self, up_prod):
        """
        A task becomes data outdated if an upstream product has a higher
        timestamp or if an upstream product is outdated
        """
        if (self.metadata.timestamp is None
                or up_prod.metadata.timestamp is None):
            return True
        else:
            return (
                (up_prod.metadata.timestamp > self.metadata.timestamp)
                # this second condition propagates outdated status
                # from indirect upstream dependencies. e.g., a -> b -> c
                # user runs in order but then it only runs a. Since a is
                # outdated, so should c
                or up_prod._is_outdated())

    def _outdated_code_dependency(self):
        """
        Determine if the product is outdated by checking the source code that
        it generated it
        """
        if self._outdated_code_dependency_status is not None:
            self.logger.debug(('Returning cached code dependencies status. '
                               'Outdated? %s'),
                              self._outdated_code_dependency_status)
            return self._outdated_code_dependency_status

        outdated, diff = self.task.dag.differ.is_different(
            a=self.metadata.stored_source_code,
            b=str(self.task.source),
            a_params=self.metadata.params,
            # process resource params to compare the file hash instead of
            # the path to the file
            b_params=process_resources(
                self.task.params.to_json_serializable(params_only=True)),
            extension=self.task.source.extension)

        self._outdated_code_dependency_status = outdated

        self.logger.debug(('Finished checking code status for task "%s" '
                           'Outdated? %s'), self.task.name,
                          self._outdated_code_dependency_status)

        if outdated:
            self.logger.info('Task "%s" has outdated code. Diff:\n%s',
                             self.task.name, diff)

        return self._outdated_code_dependency_status

    def _reset_cached_outdated_status(self):
        self._outdated_data_dependencies_status = None
        self._outdated_code_dependency_status = None
        self._is_outdated_status = None

    def __str__(self):
        return str(self._identifier)

    def __repr__(self):
        # NOTE: this assumes ._identifier has a best_repr property,
        # should we refactor it?
        return '{}({})'.format(
            type(self).__name__, self._identifier.best_repr(shorten=True))

    def __getstate__(self):
        state = self.__dict__.copy()
        # logger is not pickable, so we remove them and build
        # them again in __setstate__
        del state['logger']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.logger = logging.getLogger('{}.{}'.format(__name__,
                                                       type(self).__name__))

    def to_json_serializable(self):
        """Returns a JSON serializable version of this product
        """
        # NOTE: this is used in tasks where only JSON serializable parameters
        # are supported such as NotebookRunner that depends on papermill
        return str(self)

    def __len__(self):
        # MetaProduct return the number of products, this is a single Product
        # hence the 1
        return 1

    # Subclasses must implement the following methods

    @abc.abstractmethod
    def _init_identifier(self, identifier):
        pass  # pragma: no cover

    @abc.abstractmethod
    def fetch_metadata(self):
        pass  # pragma: no cover

    @abc.abstractmethod
    def save_metadata(self, metadata):
        pass  # pragma: no cover

    @abc.abstractmethod
    def exists(self):
        """
        This method returns True if the product exists, it is not part
        of the metadata, so there is no cached status
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def delete(self, force=False):
        """Deletes the product
        """
        pass  # pragma: no cover

    # NOTE: currently optional but there is a conflict with this. metadata
    # defines a delete public method which calls product._delete_metadata
    # but not all products implement this
    def _delete_metadata(self):
        raise NotImplementedError(
            '_delete_metadata not implemented in {}'.format(
                type(self).__name__))

    # download and upload are only relevant for File but we add them to keep
    # the API consistent

    def download(self):
        pass  # pragma: no cover

    def upload(self):
        pass  # pragma: no cover



import abc

from ploomber.constants import TaskStatus


class Executor(abc.ABC):
    """
    Abstract class for executors.

    *Executors API is subject to change*

    Executors handle task execution, they receive a dag as a parameter and
    call task.build() on each task. To iterate over tasks, it is recommended
    that they use dag.values() as it returns tasks in topological order.

    Tasks self-report its execution status (Executed/Errored) but if the
    executor runs it in a different process, it is responsible for reporting
    the status change back to the task object, as this change triggers
    changes in status in dowsmtream tasks (e.g. if a task fails, downstream
    dependencies are aborted)

    It is safe to skip task.build() on tasks that are either Skipped or
    Aborted.

    Although not strictly required, it is recommended for executors to capture
    all warnings and exceptions, then display a summary at the end of the
    execution, showing each task with their corresponding errors and warnings.

    Executors do not run hooks, these are triggered by the DAG object
    (DAG-level hooks) and Task objects (task-level hooks). These translates
    DAG hooks to be executed in the main process and Task hooks in the same
    process where Task._build() is called, this is an important detail for
    executors that run tasks in different processes, they have to report
    back the returned value from  Task._build and assign it to the
    corresponding copy of the Task in the main process. Upon sucessful
    execution, metadata is cleared up (FIXME: we should really be sending new
    metadata instead of clearing it to force a new fetch).

    Runnning tasks and task hooks in subprocesses has the advantage ensuring
    memory is cleared after the task finishes.

    To allow finishing dag.build() gracefully, executors should raise
    DAGBuildError (this will trigger the DAG.on_failure hook)


    Notes
    -----
    The following is still being defined: do we need to send the whole dag
    object? Looks like we are good by just sending the tasks
    """
    @abc.abstractmethod
    def __call__(self, dag):
        exec_status = set([t.exec_status for t in dag.values()])

        if exec_status - {
                TaskStatus.WaitingExecution,
                TaskStatus.WaitingUpstream,
                TaskStatus.Skipped,
                TaskStatus.WaitingDownload,
        }:
            raise ValueError('Tasks should only have either '
                             'TaskStatus.WaitingExecution or '
                             'TaskStatus.WaitingUpstream before attempting '
                             'to execute, got status: {}'.format(exec_status))



from ploomber.tasks import PythonCallable
from ploomber.products import EmptyProduct

# NOTE: subclassing PythonCallable would allow us to show better error messages
# when calling dag.build() on dags that should be built using InMemoryDAG


def _input_data_passer(input_data, preprocessor):
    if preprocessor is None:
        return input_data
    else:
        return preprocessor(input_data)


def input_data_passer(dag, name, preprocessor=None):
    """
    Returns a special in-memory task that forwards input data as product to
    downstream tasks.

    Parameters
    ----------
    dag : ploomber.DAG
        DAG where the task should be added

    name : str
        Task name

    preprocessor : callable, default=None
        An arbitrary callable that can be used to add custom logic to the
        input data before passing it to downstream tasks


    Returns
    -------
    PythonCallable
        A PythonCallable task with special characteristics. It cannot be
        invoked directly, but through ``InMemoryDAG(dag).build()``
    """
    return PythonCallable(_input_data_passer,
                          EmptyProduct(),
                          dag=dag,
                          name=name,
                          params={
                              'input_data': None,
                              'preprocessor': preprocessor
                          })


def in_memory_callable(callable_, dag, name, params):
    """
    Returns a special in-memory task that runs a callable with the given
    params. When calling ``InMemoryDAG(dag).build()``, this task should not
    appear in the input_data dictionary

    Parameters
    ----------
    callable_ : callable
        An arbitrary callable to execute

    dag : ploomber.DAG
        DAG where the task should be added

    name : str
        Task name

    params : dict
        Parameters to pass when calling ``callable_``. e.g passing
        ``params=dict(a=1)``, is equivalent to calling ``callable_`` with
        ``a=1``.

    Returns
    -------
    PythonCallable
        A PythonCallable task with special characteristics. It cannot be
        invoked directly, but through ``InMemoryDAG(dag).build()``
    """
    return PythonCallable(callable_,
                          EmptyProduct(),
                          dag=dag,
                          name=name,
                          params=params)



import logging
from ploomber.dag.dag import DAG
from ploomber.dag.inmemorydag import InMemoryDAG
from ploomber.dag.dagconfigurator import DAGConfigurator
from ploomber.dag.onlinedag import OnlineDAG, OnlineModel
from ploomber.env.env import Env
from ploomber.env.decorators import load_env, with_env
from ploomber.placeholders.sourceloader import SourceLoader
from ploomber.jupyter.manager import _load_jupyter_server_extension
from ploomber.util.loader import lazily_load_entry_point

__version__ = '0.19.1'

# Set default logging handler to avoid "No handler found" warnings.

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = [
    'DAG',
    'Env',
    'SourceLoader',
    'load_env',
    'with_env',
    'DAGConfigurator',
    'InMemoryDAG',
    'OnlineDAG',
    'OnlineModel',
    'lazily_load_entry_point',
]


def _jupyter_server_extension_paths():
    return [{'module': 'ploomber'}]


load_jupyter_server_extension = _load_jupyter_server_extension



from itertools import chain
from jinja2 import Environment
from jinja2.nodes import Getattr, Getitem, Assign

from ploomber.placeholders.placeholder import Placeholder


class JinjaExtractor:
    """Class for parsing jinja templates

    Parameters
    ----------
    code : str or Placeholder
        SQL code
    """
    def __init__(self, code):
        if not isinstance(code, (str, Placeholder)):
            raise TypeError('Code must be a str or Placeholder object, got: '
                            '{}'.format(type(code)))
        self.code = code
        self.ast = self._get_ast(code)

    def _get_ast(self, code):
        if isinstance(code, str):
            env = Environment()
            return env.parse(code)
        else:
            # placeholder
            env = code._template.environment
            return env.parse(code._raw)

    def get_code_as_str(self):
        if isinstance(self.code, str):
            return self.code
        else:
            return self.code._raw

    def find_variable_access(self, variable):
        """Find occurrences of {{variable.something}} and {{variable['something']}}
        """
        attr = self.ast.find_all(Getattr)
        item = self.ast.find_all(Getitem)
        return set([
            obj.arg.as_const() if isinstance(obj, Getitem) else obj.attr
            for obj in chain(attr, item) if obj.node.name == variable
        ]) or None

    def find_variable_assignment(self, variable):
        """
        Find a variable assignment: {% set variable = something %}, returns
        the node that assigns the variable
        """
        variables = {n.target.name: n.node for n in self.ast.find_all(Assign)}
        return None if variable not in variables else variables[variable]



"""
Internal utility functions. This isn't part of the public API.
"""

from ploomber.util.util import (safe_remove, image_bytes2html, isiterable,
                                requires, chdir_code)
from ploomber.util.param_grid import Interval, ParamGrid

__all__ = [
    'safe_remove',
    'image_bytes2html',
    'Interval',
    'ParamGrid',
    'isiterable',
    'requires',
    'chdir_code',
]



from ploomber.static_analysis.abc import Extractor
from ploomber.static_analysis.jinja import JinjaExtractor


class StringExtractor(Extractor):
    """
    Extract variables from a string
    """
    def extract_upstream(self):
        """Extract upstream keys used in a templated SQL script
        """
        extractor = JinjaExtractor(self.code)
        return extractor.find_variable_access(variable='upstream')

    def extract_product(self):
        raise NotImplementedError('extract_product is not implemented in '
                                  '{}'.format(type(self).__name__))



"""
Functions to obtain git repository information
"""
import json
import subprocess
import shlex
import sys
from pathlib import Path
import shutil


def _run_command(path, command):
    """Safely run command in certain path
    """
    if not Path(path).is_dir():
        raise ValueError('{} is not a directory'.format(path))

    out = subprocess.check_output(shlex.split(command), cwd=str(path))
    s = out.decode('utf-8')

    # remove trailing \n
    if s[-1:] == '\n':
        s = s[:-1]

    return s


def is_repo(path):
    """Check if the path is in a git repo"""
    if path is None:
        return False

    if not shutil.which('git'):
        return False

    out = subprocess.run(['git', '-C', str(path), 'rev-parse'],
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    repo_exists = out.returncode == 0

    if repo_exists:
        try:
            # edge case: if the repo doesn't have any commits, the following
            # will fail. we require a repo with at least one commit for git
            # to work
            git_hash(path)
        except subprocess.CalledProcessError:
            return False
        else:
            return True


def get_git_summary(path):
    """Get one line git summary: {hash} {commit-message}
    """
    return _run_command(path, 'git show --oneline -s')


def git_hash(path):
    """Get git hash

    If tag: {tag-name}
    If clean commit: {hash}
    If dirty: {hash}-dirty

    dirty: "A working tree is said to be "dirty" if it contains modifications
    which have not been committed to the current branch."
    https://mirrors.edge.kernel.org/pub/software/scm/git/docs/gitglossary.html#def_dirty
    """
    return _run_command(path, 'git describe --tags --always --dirty=-dirty')


def git_location(path):
    """
    Returns branch name if at the latest commit, otherwise the hash
    """
    hash_ = git_hash(path)
    git_branch = current_branch(path)
    return git_branch or hash_


def get_git_timestamp(path):
    """Timestamp for last commit
    """
    return int(_run_command(path, 'git log -1 --format=%ct'))


def current_branch(path):
    # seems like the most reliable way is to do:
    # git branch --show-current, but that was added in a recent git
    # version 2.22, for older versions, the one below works
    try:
        return _run_command(path, 'git symbolic-ref --short HEAD')
    except subprocess.CalledProcessError:
        # if detached head, the command above does not work, since there is
        # no current branch
        return None


def get_version(package_name):
    """Get package version
    """
    installation_path = sys.modules[package_name].__file__

    NON_EDITABLE = True if 'site-packages/' in installation_path else False

    if NON_EDITABLE:
        return getattr(package_name, '__version__')
    else:
        parent = str(Path(installation_path).parent)

    return get_git_summary(parent)


def get_diff(path):
    return _run_command(path, "git diff -- . ':(exclude)*.ipynb'")


def get_git_info(path):
    return dict(git_summary=get_git_summary(path),
                git_hash=git_hash(path),
                git_diff=get_diff(path),
                git_timestamp=get_git_timestamp(path),
                git_branch=current_branch(path),
                git_location=git_location(path))


def save_env_metadata(env, path_to_output):
    summary = get_git_summary(env.path.home)
    hash_ = git_hash(env.path.home)
    diff = get_diff(env.path.home)

    metadata = dict(summary=summary, hash=hash_)
    path_to_patch_file = Path(path_to_output).with_suffix('.patch')

    with open(path_to_output, 'w') as f:
        json.dump(metadata, f)

    with open(path_to_patch_file, 'w') as f:
        f.writelines(diff)



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# Add description here
#
[% if not is_ipynb -%]
# *Note:* You can open this file as a notebook (JupyterLab: right-click on it in the side bar -> Open With -> Notebook)
[% endif %]

# %%
# Uncomment the next two lines to enable auto reloading for imported modules
# # %load_ext autoreload
# # %autoreload 2
# For more info, see:
# https://docs.ploomber.io/en/latest/user-guide/faq_index.html#auto-reloading-code-in-jupyter

# %% tags=["parameters"]
[% if extract_upstream -%]
# If this task has dependencies, list them them here
# (e.g. upstream = ['some_task']), otherwise leave as None.
upstream = None
[% else -%]
# If this task has dependencies, declare them in the YAML spec and leave this
# as None
upstream = None
[% endif %]
[% if extract_product -%]
# product['nb'] must be the path to save the executed version of this task,
# other keys can be used to reference other output files
product = {'nb': 'products/notebook.ipynb'}
[% else -%]
# This is a placeholder, leave it as None
product = None
[% endif %]

# %%
# your code here...



import importlib
import inspect

from ploomber.sources.abc import Source
from ploomber.sources.inspect import getfile
from ploomber.util.util import signature_check
from ploomber.util.dotted_path import (load_dotted_path,
                                       lazily_locate_dotted_path)
from ploomber.static_analysis.python import PythonCallableExtractor


class CallableLoader:
    """
    If initialized with a string, it provies some functionality
    without importing it.
    """

    # TODO: this class could be used anywhere where we have to load callables
    # from dotted paths in the spec (and could also be useful for the Python
    # API): sources, hooks and clients. We can add a "lazy" option that allows
    # to keep a reference to the callable and import it until it's called
    def __init__(self, primitive, hot_reload):
        self.hot_reload = hot_reload
        self._from_dotted_path = isinstance(primitive, str)
        self._primitive = primitive

        if self.hot_reload and not self._from_dotted_path:
            self.module_name = inspect.getmodule(primitive).__name__
            self.fn_name = primitive.__name__

            # if using hot reload, we cannot keep the reference to the
            # original function, otherwise pickle will give errors
            self._primitive = None
        elif self.hot_reload and self._from_dotted_path:
            raise NotImplementedError('hot_reload is not implemented when '
                                      'initializing from a dotted path')

    def load(self):
        if self._from_dotted_path:
            return load_dotted_path(self._primitive)
        else:
            if self.hot_reload:
                module = importlib.import_module(self.module_name)
                importlib.reload(module)
                return getattr(module, self.fn_name)
            else:
                return self._primitive

    def get_source(self):
        if self._from_dotted_path:
            _, source = lazily_locate_dotted_path(self._primitive)
            return source
        else:
            return inspect.getsource(self.load())

    def get_loc(self):
        if self._from_dotted_path:
            loc, _ = lazily_locate_dotted_path(self._primitive)
            return loc
        else:
            path = getfile(self.load())
            _, line = inspect.getsourcelines(self.load())
            return '{}:{}'.format(path, line)

    @property
    def from_dotted_path(self):
        return self._from_dotted_path

    @property
    def name(self):
        if self._from_dotted_path:
            return self._primitive.split('.')[-1]
        else:
            return self.load().__name__


class PythonCallableSource(Source):
    """
    A source object to encapsulate a Python callable (i.e. functions).

    Parameters
    ----------
    primitive : callable or str
        The function to use, an be a callable object or a dotted path string

    hot_reload : bool
        If True, continuously reloads the function to have the latest version

    needs_product : bool
        Pass True if the function needs a "product" parameter to be executed.
        This is used to provide appropriate error messages when the function's
        signature does not match this argument. The primary use for this
        parameter is InMemoryDAG, since functions used as task are
        not expected to have a "product" parameter
    """
    def __init__(self, primitive, hot_reload=False, needs_product=True):
        if not (callable(primitive) or isinstance(primitive, str)):
            raise TypeError(
                f'{type(self).__name__} must be initialized '
                f'with a Python callable or str, got: {primitive!r} '
                f'(type {type(primitive).__name__})')

        self._callable_loader = CallableLoader(primitive, hot_reload)
        self._source_as_str = None
        self._loc = None
        self._hot_reload = hot_reload
        self._needs_product = needs_product

    @property
    def primitive(self):
        return self._callable_loader.load()

    def __repr__(self):
        return "{}({}) (defined at: '{}')".format(
            type(self).__name__, self.name, self.loc)

    def __str__(self):
        if self._source_as_str is None or self._hot_reload:
            self._source_as_str = self._callable_loader.get_source()

        return self._source_as_str

    @property
    def doc(self):
        return self.primitive.__doc__

    @property
    def loc(self):
        if self._loc is None or self._hot_reload:
            self._loc = self._callable_loader.get_loc()

        return self._loc

    def render(self, params):
        self._post_render_validation(None, params)

    @property
    def extension(self):
        return 'py'

    @property
    def name(self):
        return self._callable_loader.name

    def _post_render_validation(self, rendered_value, params):
        """
        Validation function executed after rendering
        """
        # TODO: we should be able to perform this validation from dotted
        # paths as well
        if not self._callable_loader.from_dotted_path:
            to_validate = set(params)
            fn_params = inspect.signature(self.primitive).parameters

            if not self._needs_product and 'product' in fn_params:
                raise TypeError(
                    f'Function {self.name!r} should not have '
                    'a \'product\' parameter, but return its result instead')

            # if source does not need product to be called and we got
            # a product parameter, remove it. NOTE: Task.render removes
            # product from params if dealing with an EmptyProduct - it's
            # better to always pass it and just remove it here to avoid
            # the second condition. i don't think there is any other
            # condition were we don't receive product here
            if not self._needs_product and 'product' in to_validate:
                to_validate.remove('product')

            signature_check(self.primitive, to_validate, self.name)

    def _post_init_validation(self, value):
        # TODO: verify the callable has a product parameter, if it's required
        pass

    @property
    def variables(self):
        raise NotImplementedError

    def extract_upstream(self):
        return PythonCallableExtractor(str(self)).extract_upstream()

    def extract_product(self):
        return PythonCallableExtractor(str(self)).extract_product()



from ploomber.io.file import CSVIO, ParquetIO
from ploomber.io.terminalwriter import TerminalWriter
from ploomber.io.loaders import FileLoaderMixin
from ploomber.io.serialize import serializer, serializer_pickle
from ploomber.io.unserialize import unserializer, unserializer_pickle

__all__ = [
    'CSVIO',
    'ParquetIO',
    'TerminalWriter',
    'FileLoaderMixin',
    'serializer',
    'serializer_pickle',
    'unserializer',
    'unserializer_pickle',
]



"""
Clients that communicate with shell processes
"""
import os
import tempfile
from pathlib import Path
import random
import string
import logging
import shlex
import subprocess

from ploomber.clients.client import Client
from ploomber.placeholders.placeholder import Placeholder
from ploomber.util import requires


class ShellClient(Client):
    """Client to run command in the local shell

    Parameters
    ----------
    run_template : str
        Template for executing commands, must include the {{path_to_code}}
        placeholder which will point to the rendered code. Defaults to
        'bash {{path_to_code}}'
    subprocess_run_kwargs : dict
        Keyword arguments to pass to the subprocess.run when running
        run_template
    """
    def __init__(self,
                 run_template='bash {{path_to_code}}',
                 subprocess_run_kwargs={
                     'stderr': subprocess.PIPE,
                     'stdout': subprocess.PIPE,
                     'shell': False
                 }):
        self.subprocess_run_kwargs = subprocess_run_kwargs
        self.run_template = run_template
        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))

    @property
    def connection(self):
        raise NotImplementedError('ShellClient does not need a connection')

    def execute(self, code):
        """Run code
        """
        fd, path_to_tmp = tempfile.mkstemp()
        os.close(fd)
        Path(path_to_tmp).write_text(code)

        run_template = Placeholder(self.run_template)
        # we need quoting to make windows paths keep their \\ separators when
        # running shlex.split
        source = run_template.render(
            dict(path_to_code=shlex.quote(path_to_tmp)))

        res = subprocess.run(shlex.split(source), **self.subprocess_run_kwargs)
        Path(path_to_tmp).unlink()

        stdout = res.stdout.decode('utf-8')
        stderr = res.stderr.decode('utf-8')

        if res.returncode != 0:
            # log source code without expanded params
            self._logger.info(
                ('{} returned stdout: '
                 '{}\nstderr: {}\n'
                 'exit status {}').format(code, stdout, stderr,
                                          res.returncode))
            raise RuntimeError(
                ('Error executing code.\nReturned stdout: '
                 '{}\nstderr: {}\n'
                 'exit status {}').format(stdout, stderr, res.returncode))
        else:
            self._logger.info(('Finished running {}. stdout: {},'
                               ' stderr: {}').format(self, stdout, stderr))

    def close(self):
        pass


class RemoteShellClient(Client):
    """EXPERIMENTAL: Client to run commands in a remote shell
    """
    @requires(['paramiko'], 'RemoteShellClient')
    def __init__(self,
                 connect_kwargs,
                 path_to_directory,
                 run_template='bash {{path_to_code}}'):
        """

        path_to_directory: str
            A path to save temporary files

        connect_kwargs: dict
            Parameters to send to the paramiko.SSHClient.connect constructor
        """
        self.path_to_directory = path_to_directory
        self.connect_kwargs = connect_kwargs
        self.run_template = run_template
        self._raw_client = None
        self._logger = logging.getLogger('{}.{}'.format(
            __name__,
            type(self).__name__))

    @property
    def connection(self):
        # client has not been created
        if self._raw_client is None:
            self._raw_client, policy = ssh_client_and_policy()
            self._raw_client.set_missing_host_key_policy(policy)
            self._raw_client.connect(**self.connect_kwargs)

        # client has been created but still have to check if it's active:
        else:

            is_active = False

            # this might not always work: https://stackoverflow.com/a/28288598
            if self._raw_client.get_transport() is not None:
                is_active = self._raw_client.get_transport().is_active()

            if not is_active:
                self._raw_client.connect(**self.connect_kwargs)

        return self._raw_client

    def _random_name(self):
        filename = (''.join(
            random.choice(string.ascii_letters) for i in range(16)))
        return filename

    def read_file(self, path):
        """Read file in remote path
        """
        ftp = self.connection.open_sftp()

        # tmp location for file
        fd, path_to_tmp = tempfile.mkstemp()
        os.close(fd)
        ftp.get(path, path_to_tmp)

        path_to_tmp = Path(path_to_tmp)
        content = path_to_tmp.read_text()
        path_to_tmp.unlink()

        ftp.close()

        return content

    def write_to_file(self, content, path):
        """Write file to remote path
        """
        ftp = self.connection.open_sftp()

        fd, path_to_tmp = tempfile.mkstemp()
        os.close(fd)
        path_to_tmp = Path(path_to_tmp)
        path_to_tmp.write_text(content)
        ftp.put(path_to_tmp, path)

        ftp.close()
        path_to_tmp.unlink()

    def execute(self, code):
        """Run code
        """
        ftp = self.connection.open_sftp()
        path_remote = self.path_to_directory + self._random_name()

        fd, path_to_tmp = tempfile.mkstemp()
        os.close(fd)
        Path(path_to_tmp).write_text(code)

        ftp.put(path_to_tmp, path_remote)
        ftp.close()

        run_template = Placeholder(self.run_template)
        source = run_template.render(dict(path_to_code=path_remote))

        # stream stdout. related: https://stackoverflow.com/q/31834743
        # using pty is not ideal, fabric has a clean implementation for this
        # worth checking out
        stdin, stdout, stderr = self.connection.exec_command(source,
                                                             get_pty=True)

        for line in iter(stdout.readline, ""):
            self._logger.info('(STDOUT): {}'.format(line))

        returncode = stdout.channel.recv_exit_status()

        stdout = ''.join(stdout)
        stderr = ''.join(stderr)

        if returncode != 0:
            # log source code without expanded params
            self._logger.info(
                '%s returned stdout: '
                '%s and stderr: %s '
                'and exit status %s', code, stdout, stderr, returncode)
            raise subprocess.CalledProcessError(returncode, code)
        else:
            self._logger.info('Finished running %s. stdout: %s,'
                              ' stderr: %s', self, stdout, stderr)

        return {'returncode': returncode, 'stdout': stdout, 'stderr': stderr}

    def close(self):
        if self._raw_client is not None:
            self._logger.info('Closing client %s', self._raw_client)
            self._raw_client.close()
            self._raw_client = None


def ssh_client_and_policy():
    # lazily loading paramiko, importing it takes a bit and we don't want
    # to slow down importing ploomber.clients. Instead of having this code
    # in the client's constructor, we add this function to helps us mock the
    # for testing
    import paramiko
    return paramiko.SSHClient(), paramiko.AutoAddPolicy()



from pathlib import PurePosixPath
from urllib import parse
import cgi
import sys
from pathlib import Path
from urllib.request import urlretrieve, urlopen
from urllib.error import HTTPError
from concurrent.futures import ThreadPoolExecutor, as_completed
import os
from glob import glob
import zipfile
from functools import wraps
from datetime import datetime
import json

import click
import humanize

from ploomber.table import Table
from ploomber.cloud import io, config
from ploomber.exceptions import BaseException
from ploomber.spec import DAGSpec
from ploomber.dag import util
from ploomber.cli.cloud import get_key
from ploomber import _requests

HOST = os.environ.get('PLOOMBER_CLOUD_HOST', 'https://api.ploomber.io')


def _remove_prefix(path):
    parts = PurePosixPath(path).parts[2:]
    return str(PurePosixPath(*parts))


def _download_file(url, skip_if_exists=False, raise_on_missing=False):
    try:
        remotefile = urlopen(url)
    except HTTPError as e:
        if e.code == 404:
            path = _remove_prefix(parse.urlparse(url).path[1:])

            if raise_on_missing:
                raise FileNotFoundError(
                    'The requested file does not exist.'
                    ' Upload it to cloud storage and try again.')
            else:
                click.echo(f'File not found: {path}')
                return

        else:
            raise

    content_disposition = remotefile.info()['Content-Disposition']

    if content_disposition:
        _, params = cgi.parse_header(content_disposition)
        path = params["filename"]
    else:
        path = Path(parse.urlparse(url).path[1:])

    if skip_if_exists and Path(path).exists():
        print(f'{path} exists, skipping...')
    else:
        Path(path).parent.mkdir(exist_ok=True, parents=True)
        print(f'Downloading {path}')
        urlretrieve(url, path)

    return path


def download_from_presigned(presigned):
    if not presigned:
        click.echo('No files matched the criteria.\n'
                   'To list files: ploomber cloud products')
        return

    with ThreadPoolExecutor(max_workers=64) as executor:
        future2url = {
            executor.submit(_download_file, url=url): url
            for url in presigned
        }

        for future in as_completed(future2url):
            exception = future.exception()

            if exception:
                task = future2url[future]
                raise RuntimeError(
                    'An error occurred when downloading product from '
                    f'url: {task!r}') from exception


def auth_header(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        api_key = get_key()

        if api_key:
            headers = {
                "Authorization": api_key,
                "Content-Type": "application/json"
            }

            return func(headers, *args, **kwargs)
        else:
            click.secho('Error: Missing api key', fg='red')
            sys.exit(1)

    return wrapper


@auth_header
def runs_new(headers, metadata):
    """Register a new run in the database
    """
    response = _requests.post(f"{HOST}/runs", headers=headers, json=metadata)
    return response.json()['runid']


@auth_header
def runs_update(headers, runid, graph):
    """Update run status, store graph
    """
    return _requests.put(f"{HOST}/runs/{runid}", headers=headers,
                         json=graph).json()


@auth_header
def runs_register_ids(headers, runid, ids):
    """Update run status, store ids
    """
    return _requests.put(f"{HOST}/runs/{runid}/ids", headers=headers,
                         json=ids).json()


def _parse_datetime(timestamp):
    try:
        return humanize.naturaltime(
            datetime.fromisoformat(timestamp),
            when=datetime.utcnow(),
        )
    # Python 3.6 does not have fromisoformat
    except AttributeError:
        return timestamp


@auth_header
def runs(headers):
    res = _requests.get(f"{HOST}/runs", headers=headers).json()

    for run in res:
        run['created_at'] = _parse_datetime(run['created_at'])

    print(Table.from_dicts(res))


# NOTE: this doesn't need authentication (add unit test)
def tasks_update(task_id, status):
    return _requests.get(f"{HOST}/tasks/{task_id}/{status}").json()


@auth_header
def run_detail(headers, run_id):
    res = _requests.get(f"{HOST}/runs/{run_id}", headers=headers).json()
    return res


def run_detail_print(run_id):
    out = run_detail(run_id)
    tasks = out['tasks']
    run = out['run']

    if run['status'] == 'created':
        click.echo('Run created...')
    elif tasks:
        click.echo(Table.from_dicts(tasks))
    else:
        click.echo('Pipeline up-to-date, no tasks scheduled for this run.')

    return out


@auth_header
def run_logs(headers, run_id):
    res = _requests.get(f"{HOST}/runs/{run_id}/logs", headers=headers).json()

    for name, log in res.items():
        click.echo(f'\n\n***** START OF LOGS FOR TASK: {name} *****')
        click.echo(log)
        click.echo(f'***** END OF LOGS FOR TASK: {name} *****')


@auth_header
def run_logs_image(headers, run_id, tail=None):
    res = _requests.get(f"{HOST}/runs/{run_id}/logs/image", headers=headers)

    if not len(res.text):
        out = "Image build hasn't started yet..."
    elif tail:
        out = '\n'.join(res.text.splitlines()[-tail:])
    else:
        out = res.text

    click.echo(out)


@auth_header
def run_abort(headers, run_id):
    _requests.get(f"{HOST}/runs/{run_id}/abort", headers=headers).json()
    print("Aborted.")


@auth_header
def run_finished(headers, runid):
    response = _requests.get(f"{HOST}/runs/{runid}/finished", headers=headers)
    return response


@auth_header
def products_list(headers):
    res = _requests.get(f"{HOST}/products", headers=headers).json()

    if res:
        print(Table.from_dicts([{'path': r} for r in res]))
    else:
        print("No products found.")


@auth_header
def data_list(headers):
    res = _requests.get(f"{HOST}/data", headers=headers).json()

    if res:
        print(Table.from_dicts([{'path': r} for r in res]))
    else:
        print("No data found.")


@auth_header
def products_download(headers, pattern):
    res = _requests.post(f"{HOST}/products",
                         headers=headers,
                         json=dict(pattern=pattern)).json()
    download_from_presigned(res)


def _has_prefix(path, prefixes):
    if not prefixes:
        return False

    return any(str(path).startswith(prefix) for prefix in prefixes)


def zip_project(force, runid, github_number, verbose, ignore_prefixes=None):
    ignore_prefixes = ignore_prefixes or []

    if Path("project.zip").exists():
        if verbose:
            click.secho("Deleting existing project.zip...", fg="yellow")
        Path("project.zip").unlink()

    files = glob("**/*", recursive=True)

    # TODO: ignore __pycache__, ignore .git directory
    with zipfile.ZipFile("project.zip", "w", zipfile.ZIP_DEFLATED) as zip:
        for path in files:
            if not _has_prefix(path, ignore_prefixes):
                zip.write(path, arcname=path)
            else:
                click.echo(f'Ignoring: {path}')

        # NOTE: it's weird that force is loaded from the buildspec but the
        # other two parameters are actually loaded from dynamodb
        zip.writestr(
            '.ploomber-cloud',
            json.dumps({
                'force': force,
                'runid': runid,
                'github_number': github_number
            }))

    MAX = 5 * 1024 * 1024

    if Path('project.zip').stat().st_size > MAX:
        raise BaseException("Error: Your project's source code is over "
                            "5MB, which isn't supported. Tip: Ensure there "
                            "aren't any large data files and try again")


@auth_header
def get_presigned_link(headers):
    return _requests.get(f"{HOST}/upload", headers=headers).json()


def upload_zipped_project(response, verbose):
    with open("project.zip", "rb") as f:
        files = {"file": f}
        http_response = _requests.post(response["url"],
                                       data=response["fields"],
                                       files=files)

    if http_response.status_code != 204:
        raise ValueError(f"An error happened: {http_response}")

    if verbose:
        click.secho("Uploaded project, starting execution...", fg="green")


@auth_header
def trigger(headers):
    res = _requests.get(f"{HOST}/trigger", headers=headers).json()
    return res


def upload_project(force=False,
                   github_number=None,
                   github_owner=None,
                   github_repo=None,
                   verbose=False):
    dag = DAGSpec('pipeline.yaml').to_dag().render(show_progress=False)

    # TODO: test
    if not Path("requirements.lock.txt").exists():
        raise BaseException("Missing requirements.lock.txt file, add one "
                            "with the dependencies to install")

    config.validate()

    runid = runs_new(
        dict(force=force,
             github_number=github_number,
             github_owner=github_owner,
             github_repo=github_repo))

    # TODO: ignore relative paths in products
    if verbose:
        click.echo("Zipping project -> project.zip")

    # TODO: test
    zip_project(force,
                runid,
                github_number,
                verbose,
                ignore_prefixes=util.extract_product_prefixes(dag))

    if verbose:
        click.echo("Uploading project...")

    response = get_presigned_link()

    upload_zipped_project(response, verbose)

    if verbose:
        click.echo("Starting build...")

    trigger()

    # TODO: if anything fails after runs_new, update the status to error
    # convert runs_new into a context manager

    return runid


@auth_header
def upload_data(headers, path):
    key = Path(path).name

    create = _requests.post(f"{HOST}/upload/data/create",
                            headers=headers,
                            json=dict(key=key,
                                      n_parts=io.n_parts(path))).json()

    gen = io.UploadJobGenerator(path,
                                key=key,
                                upload_id=create['upload_id'],
                                links=create['urls'])

    click.echo('Uploading...')
    parts = gen.upload()

    _requests.post(f"{HOST}/upload/data/complete",
                   headers=headers,
                   json=dict(key=key,
                             parts=parts,
                             upload_id=create['upload_id'])).json()

    # print snippet showing how to download it in the pipeline


@auth_header
def download_data(headers, key):
    response = _requests.post(f"{HOST}/download/data",
                              headers=headers,
                              json=dict(key=key))
    return response


@auth_header
def delete_data(headers, pattern):
    response = _requests.delete(f"{HOST}/data",
                                headers=headers,
                                json=dict(pattern=pattern))
    print(response.json())


@auth_header
def delete_products(headers, pattern):
    response = _requests.delete(f"{HOST}/products",
                                headers=headers,
                                json=dict(pattern=pattern))
    print(response.json())
